#ifndef OVM_HPP
#define OVM_HPP

#include <map>
#include <vector>
#include <complex>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <limits.h>
#include <math.h>

#include <time.h>

#define TEAM_ID 652
#define CAFEBABE 0xCAFEBABE

#define MAX_TARGETS 11
#define FATE_ID (-1)
#define STATION_ID (-2)

const double G = 6.67428e-11;
const double Re = 6.357e6;
const double Me = 6.0e24;
const double Mm = 7.347e22;

void vm_run_1(int scenario_id, double* mem, double* input, double* output, int& status);
void vm_run_2(int scenario_id, double* mem, double* input, double* output, int& status);
void vm_run_3(int scenario_id, double* mem, double* input, double* output, int& status);
void vm_run_4(int scenario_id, double* mem, double* input, double* output, int& status);
extern double vm_init_memmap_1[];
extern double vm_init_memmap_2[];
extern double vm_init_memmap_3[];
extern double vm_init_memmap_4[];
extern int vm_memsize_1;
extern int vm_memsize_2;
extern int vm_memsize_3;
extern int vm_memsize_4;

typedef std::complex<double> P;


class OVM {

private:

    int scenario_id;
    int problem_id;
    int instance_id;
    double input[0x4000];
    double output[0x4000];
    double mem[0x4000];
    int status;
    int timestep;
    P last_thrust;
    bool thrust_hooked;
    double score;
    double fuel;
    std::map< int, std::map<int, double> > log;
    P current_fate, previous_fate;
    P current_moon, previous_moon;
    P current_station, previous_station;
    P current_targets[MAX_TARGETS], previous_targets[MAX_TARGETS];
    int target_port_base;
    bool auto_save;
    int auto_vis;
    bool safe_mode;
    bool anti_moon;


public:

    OVM() {
        auto_save = false;
        auto_vis = -1;
        safe_mode = true;
        anti_moon = false;
        scenario_id = 0;
    }

    explicit OVM(int scenario_id) {
        auto_vis = -1;
        auto_save = false;
        safe_mode = true;
        anti_moon = false;
        init(scenario_id);
    }

    OVM(const OVM& rhs) {
        auto_vis = -1;
        auto_save = false;
        safe_mode = false; // コピーされてできたOVMではセーフモードを使用しない
        anti_moon = false;
        *this = rhs;
    }

    OVM& operator=(const OVM& rhs) {
        scenario_id = rhs.scenario_id;
        problem_id = rhs.problem_id;
        instance_id = rhs.instance_id;

	// Copy input.
        // memcpy(input, rhs.input, sizeof(input));
	input[2] = rhs.input[2];
	input[3] = rhs.input[3];
        input[0x3e80] = (double)scenario_id;

	// Copy output.
        // memcpy(output, rhs.output, sizeof(output));
	memcpy(output, rhs.output, sizeof(double)*0x66);

	// Copy mem.
	// memcpy(mem, rhs.mem, sizeof(mem));
        switch(problem_id) {
        case 1: memcpy(mem, rhs.mem, sizeof(double)*vm_memsize_1); break;
        case 2: memcpy(mem, rhs.mem, sizeof(double)*vm_memsize_2); break;
        case 3: memcpy(mem, rhs.mem, sizeof(double)*vm_memsize_3); break;
        case 4: memcpy(mem, rhs.mem, sizeof(double)*vm_memsize_4); break;
        }

        status = rhs.status;
        timestep = rhs.timestep;
        last_thrust = rhs.last_thrust;
        thrust_hooked = rhs.thrust_hooked;
        score = rhs.score;
        fuel = rhs.score;
        log = rhs.log;
        current_fate = rhs.current_fate;
        previous_fate = rhs.previous_fate;
        current_moon = rhs.current_moon;
        previous_moon = rhs.previous_moon;
        current_station = rhs.current_station;
        previous_station = rhs.previous_station;
        for(int i = 0; i < MAX_TARGETS; i++) {
            current_targets[i] = rhs.current_targets[i];
            previous_targets[i] = rhs.previous_targets[i];
        }
        target_port_base = rhs.target_port_base;
        // auto_saveはコピーしない
        // auto_visはコピーしない
        anti_moon = rhs.anti_moon;
        return* this;
    }

    virtual ~OVM() {
        uninit();
    }

    void uninit() {
        scenario_id = 0;
        log.clear();
    }

    void init() {
        init(scenario_id);
    }

    void init(int scenario_id) {
        uninit();
        this->scenario_id = scenario_id;
        problem_id = scenario_id / 1000;
        instance_id = scenario_id % 10;
        assert(problem_id * 1000 + instance_id == scenario_id);
        assert(1 <= problem_id && problem_id <= 4 && 1 <= instance_id && instance_id <= 4);
        switch(problem_id) {
        case 1:
        case 2:
        case 3:
            target_port_base = 4;
            break;
        case 4:
            target_port_base = 7;
            break;
        }
        memset(mem, 0, sizeof(mem));
        memset(input, 0, sizeof(input));
        memset(output, 0, sizeof(output));
        switch(problem_id) {
        case 1: memcpy(mem, vm_init_memmap_1, sizeof(double)*vm_memsize_1); break;
        case 2: memcpy(mem, vm_init_memmap_2, sizeof(double)*vm_memsize_2); break;
        case 3: memcpy(mem, vm_init_memmap_3, sizeof(double)*vm_memsize_3); break;
        case 4: memcpy(mem, vm_init_memmap_4, sizeof(double)*vm_memsize_4); break;
        }
        status = 0;
        timestep = 0;
        last_thrust = P();
        thrust_hooked = false;
        score = 0;
        log.clear();
        input[0x3e80] = (double)scenario_id;
        push_log(0x3e80, (double)scenario_id);
        push_log(2, 0);
        push_log(3, 0);
    }

    bool run_free() {
        return run(P(), true);
    }
    bool run(P thrust = P(), bool freedom = false) {
        assert(scenario_id);
        if (!freedom && anti_moon && timestep >= 1) {
            double Dm = abs(current_fate-current_moon);
            P anti_moon_thrust = (current_fate-current_moon)*G*Mm/(Dm*Dm*Dm);
            thrust += anti_moon_thrust;
        }
        if (timestep >= 1 && abs(thrust) > current_fuel() && safe_mode) {
            // 燃料切れになるくらいならやめてやる！
            printf("ovm: running out of fuel, hibernating... (it may take minutes)\n");
            if (score == 0)
                while(run_free());
            if (score > 0) {
                printf("ovm: crashed!\n");
                print_status();
            }
            else {
                printf("ovm: finished with %.3f pts.\n", current_score());
                if (auto_save) {
                    printf("ovm.auto_save: saved.\n");
                    save_log();
                }
            }
            exit(0);
        }
        if ((thrust_hooked && thrust != P()) || (auto_vis > 0 && (timestep % auto_vis == 0 || thrust != P()) && timestep >= 2))
            print_vis(thrust, false);
        thrust_hooked = false;
        // あまり変わっていないなら無視
        if (abs(thrust-last_thrust) < 1e-6)
            thrust = last_thrust;
        input[2] = thrust.real();
        input[3] = thrust.imag();
        if (thrust.real() != last_thrust.real())
            push_log(2, thrust.real());
        if (thrust.imag() != last_thrust.imag())
            push_log(3, thrust.imag());
        last_thrust = thrust;
        switch(problem_id) {
        case 1: vm_run_1(scenario_id, mem, input, output, status); break;
        case 2: vm_run_2(scenario_id, mem, input, output, status); break;
        case 3: vm_run_3(scenario_id, mem, input, output, status); break;
        case 4: vm_run_4(scenario_id, mem, input, output, status); break;
        }
        timestep++;
        fuel = output[1];
        double new_score = output[0];
        if (score == 0 && new_score != 0) {
            log[timestep];
            score = new_score;
            if (auto_save) {
                printf("ovm.auto_save: finished with %.3f pts.\n", new_score);
                if (score > 0) {
                    printf("ovm.auto_save: saved.\n");
                    save_log();
                }
            }
        }
        update_positions();
        return (score == 0);
    }

    void set_auto_save(bool flag) {
        auto_save = flag;
    }

    void set_auto_vis(int interval) {
        auto_vis = interval;
    }

    void set_safe_mode(bool flag) {
        safe_mode = flag;
    }

    void set_anti_moon(bool flag) {
        anti_moon = flag;
    }

    bool finished() {
        return (score != 0);
    }

    int current_timestep() {
        assert(scenario_id);
        return timestep;
    }

    double current_score() {
        return score;
    }

    double current_fuel() {
        return fuel;
    }

    double target_orbit_radius() {
        return output[4];
    }

    P current_fate_position() {
        return current_fate;
    }
    P previous_fate_position() {
        return previous_fate;
    }
    P current_fate_velocity() {
        return velocity_from_delta(previous_fate, current_fate, last_thrust);
    }

    P current_moon_position() {
        return current_moon;
    }
    P previous_moon_position() {
        return previous_moon;
    }
    P current_moon_velocity() {
        return velocity_from_delta(previous_moon, current_moon, P(), false);
    }

    P current_station_position() {
        return current_station;
    }
    P previous_station_position() {
        return previous_station;
    }
    P current_station_velocity() {
        return velocity_from_delta(previous_station, current_station, P());
    }

    P current_target_position(int n = 0) {
        if (n == STATION_ID)
            return current_station_position();
        return current_targets[n];
    }
    P previous_target_position(int n = 0) {
        if (n == STATION_ID)
            return previous_station_position();
        return previous_targets[n];
    }
    P current_target_velocity(int n = 0) {
        if (n == STATION_ID)
            return current_station_velocity();
        return velocity_from_delta(previous_targets[n], current_targets[n], P());
    }

    bool has_target() {
        assert(scenario_id);
        return (problem_id >= 2);
    }
    bool has_moon() {
        assert(scenario_id);
        return (problem_id == 4);
    }

    bool target_collected(int n) {
        assert(scenario_id);
        if (n == STATION_ID)
            return (abs(current_fate-current_station) < 1e3);
        if (problem_id == 4)
            return (output[target_port_base+3*n+2] > 0.5);
        return (score > 0);
    }

    void print_status(int target_id = FATE_ID) {
        printf("%7d: fate=(%10.0f, %10.0f)|%5.0f|{%5.4f, %5.4f}|%5.4f|, fuel=%5.4f",
               current_timestep(),
               current_fate_position().real(), current_fate_position().imag(),
               abs(current_fate_position()),
               current_fate_velocity().real(), current_fate_velocity().imag(),
               abs(current_fate_velocity()), current_fuel());
        if (problem_id == 2 || problem_id == 3) {
            printf(", target=(%10.0f, %10.0f)|%5.0f|{%5.4f, %5.4f}|%5.4f|, dist=%10.0f",
                   current_target_position().real(), current_target_position().imag(),
                   abs(current_target_position()),
                   current_target_velocity().real(), current_target_velocity().imag(),
                   abs(current_target_velocity()),
                   abs(current_target_position()-current_fate_position()));
        }
        else if (problem_id == 4) {
            if (target_id != FATE_ID) {
                printf(", target=(%10.0f, %10.0f)|%5.0f|{%5.4f, %5.4f}|%5.4f|, dist=%10.0f",
                       current_target_position(target_id).real(), current_target_position(target_id).imag(),
                       abs(current_target_position(target_id)),
                       current_target_velocity(target_id).real(), current_target_velocity(target_id).imag(),
                       abs(current_target_velocity(target_id)),
                       abs(current_target_position(target_id)-current_fate_position()));
            }
            int collected_targets = 0;
            for(int i = 0; i < MAX_TARGETS; i++)
                if (target_collected(i))
                    collected_targets++;
            printf(", collected targets = %d", collected_targets);
        }
        printf("\n");
    }

    void print_vis(P thrust = P(), bool fate_only = false) {
        thrust_hooked = (thrust == P());
        printf("VIS %d fate %.0f %.0f %.4f %.4f\n",
               current_timestep(),
               current_fate_position().real(), current_fate_position().imag(),
               thrust.real(), thrust.imag());
        if (fate_only)
            return;
        if (problem_id == 2 || problem_id == 3) {
            printf("VIS %d target %.0f %.0f 0 0\n",
                   current_timestep(),
                   current_target_position().real(), current_target_position().imag());
        }
        else if (problem_id == 4) {
            for(int i = 0; i < MAX_TARGETS; i++) {
                printf("VIS %d target%d %.0f %.0f 0 0\n",
                       current_timestep(), i,
                       current_target_position(i).real(), current_target_position(i).imag());
            }
            printf("VIS %d station %.0f %.0f 0 0\n",
                   current_timestep(),
                   current_station_position().real(), current_station_position().imag());
            printf("VIS %d moon %.0f %.0f 0 0\n",
                   current_timestep(),
                   current_moon_position().real(), current_moon_position().imag());
        }
    }

    void save_log() {
        static char filename[256];
        static char date[64];
        struct tm* now;
        time_t t = time(NULL);
        now = localtime(&t);
        strftime(date, sizeof(date), "%Y%m%d%H%M%S", now);
        sprintf(filename, "%d-%s-%.0fpts.osf", scenario_id, date, current_score());
        save_log(filename);
    }

    void save_log(const char* filename) {
        assert(scenario_id);

        FILE* fp = fopen(filename, "w");
        assert(fp);

        unsigned int tmp_int;
        double tmp_double;
        tmp_int = CAFEBABE; fwrite(&tmp_int, sizeof(unsigned int), 1, fp);
        tmp_int = TEAM_ID; fwrite(&tmp_int, sizeof(unsigned int), 1, fp);
        tmp_int = scenario_id; fwrite(&tmp_int, sizeof(unsigned int), 1, fp);

        for(std::map<int, std::map<int, double> >::iterator it = log.begin(); it != log.end(); ++it) {
            tmp_int = it->first; fwrite(&tmp_int, sizeof(unsigned int), 1, fp);
            tmp_int = it->second.size(); fwrite(&tmp_int, sizeof(unsigned int), 1, fp);
            for(std::map<int, double>::iterator jt = it->second.begin(); jt != it->second.end(); ++jt) {
                tmp_int = jt->first; fwrite(&tmp_int, sizeof(unsigned int), 1, fp);
                tmp_double = jt->second; fwrite(&tmp_double, sizeof(double), 1, fp);
            }
        }
        fclose(fp);
    }


private:

    P velocity_from_delta(P previous_position, P current_position, P last_thrust, bool consider_moon = true) {
        assert(scenario_id);
        double previous_height_earth = abs(previous_position);
        double current_height_earth = abs(current_position);
        P previous_gravity = -G * Me * previous_position / (previous_height_earth*previous_height_earth*previous_height_earth);
        P current_gravity = -G * Me * current_position / (current_height_earth*current_height_earth*current_height_earth);
        if (problem_id == 4 && consider_moon) {
            P previous_relative_moon = previous_position - previous_moon;
            double previous_height_moon = abs(previous_relative_moon);
            P current_relative_moon = current_position - current_moon;
            double current_height_moon = abs(current_relative_moon);
            previous_gravity -= G * Mm * previous_relative_moon / (previous_height_moon*previous_height_moon*previous_height_moon);
            current_gravity -= G * Mm * current_relative_moon / (current_height_moon*current_height_moon*current_height_moon);
        }
        P last_velocity = (current_position - previous_position) - 0.5 * (previous_gravity + last_thrust);
        P current_velocity = last_velocity + last_thrust + 0.5 * (previous_gravity + current_gravity);
        return current_velocity;
    }

    void update_positions() {
        previous_fate = current_fate;
        previous_moon = current_moon;
        previous_station = current_station;
        for(int i = 0; i < MAX_TARGETS; i++)
            previous_targets[i] = current_targets[i];
        current_fate = -P(output[2], output[3]);
        current_moon = current_fate + P(output[0x64], output[0x65]);
        current_station = current_fate + P(output[4], output[5]);
        for(int i = 0; i < MAX_TARGETS; i++)
            current_targets[i] = current_fate + P(output[target_port_base+3*i+0], output[target_port_base+3*i+1]);
    }

    void push_log(int addr, double value) {
        assert(scenario_id);
        if (score == 0)
            log[timestep][addr] = value;
    }

};


#endif
