#include <stdio.h>
#include <stdlib.h>

#define MEMSIZE (1<<14)

double memory[MEMSIZE];

const char* OPS[5] = { "<", "<=", "==", ">=", ">" };


void emit_instruction(unsigned int code, unsigned int pos) {
    unsigned int op = code >> 28;
    if (op == 0) {
        // S-Type
        unsigned int imm = (code >> 21) & 0x7;
        unsigned int r1 = code & 0x3fff;
        op = code >> 24;
        switch(op) {
        case 0: // Noop
            printf("// noop\n");
            break;
        case 1: // Cmpz
            if (imm > 4) {
                fprintf(stderr, "ERROR: unknown instruction %08x\n", code);
                exit(1);
            }
            printf("status = (mem[%u] %s 0);\n", r1, OPS[imm]);
            break;
        case 2: // Sqrt
            printf("mem[%u] = sqrt(fabs(mem[%u]));\n", pos, r1);
            break;
        case 3: // Copy
            printf("mem[%u] = mem[%u];\n", pos, r1);
            break;
        case 4: // Input
            printf("mem[%u] = input[%u];\n", pos, r1);
            break;
        default:
            fprintf(stderr, "ERROR: unknown instruction %08x\n", code);
            exit(1);
        }
    }
    else {
        // D-Type
        unsigned int r1 = (code >> 14) & 0x3fff;
        unsigned int r2 = code & 0x3fff;
        switch(op) {
        case 1: // Add
            printf("mem[%u] = mem[%u] + mem[%u];\n", pos, r1, r2);
            break;
        case 2: // Sub
            printf("mem[%u] = mem[%u] - mem[%u];\n", pos, r1, r2);
            break;
        case 3: // Mult
            printf("mem[%u] = mem[%u] * mem[%u];\n", pos, r1, r2);
            break;
        case 4: // Div
            printf("if (mem[%u] == 0) mem[%u] = 0; else mem[%u] = mem[%u] / mem[%u];\n", r2, pos, pos, r1, r2);
            break;
        case 5: // Output
            printf("output[%u] = mem[%u];\n", r1, r2);
            break;
        case 6: // Phi
            printf("mem[%u] = mem[status ? %u : %u];\n", pos, r1, r2);
            break;
        default:
            fprintf(stderr, "ERROR: unknown instruction %08x\n", code);
            exit(1);
        }
    }
}

int main(void) {
    unsigned int pos = 0, end;

    printf("#include <math.h>\n"
           "#include <string.h>\n"
           "#include <stdio.h>\n"
           "#include <stdlib.h>\n"
           "const double G = 6.67428e-11, Re = 6.357e6, Me = 6.0e24;\n"
           "#define MEMSIZE %u\n"
           "static double mem[MEMSIZE], input[MEMSIZE], output[MEMSIZE];\n"
           "static double save_mem[MEMSIZE], save_input[MEMSIZE], save_output[MEMSIZE];\n"
           "static int status, save_status;\n"
           "static FILE* flog, *save_flog;\n"
           "static int timestep, save_timestep;\n"
           "static double lx, ly, vx, vy;\n"
           "static double save_lx, save_ly, save_vx, save_vy;\n"
           "double vm_run() {\n",
           MEMSIZE);

    for(;;) {
        unsigned int code;

        fread(&memory[pos], sizeof(double), 1, stdin);
        fread(&code, sizeof(unsigned int), 1, stdin);
        if (feof(stdin))
            break;
        emit_instruction(code, pos);
        pos++;

        fread(&code, sizeof(unsigned int), 1, stdin);
        fread(&memory[pos], sizeof(double), 1, stdin);
        if (feof(stdin))
            break;
        emit_instruction(code, pos);
        pos++;
    }
    end = pos;

    printf("timestep++;\n"
           "{\n"
           "double ax = input[2], ay = input[3], cx = output[2], cy = output[3];\n"
           "double cr = sqrt(cx*cx+cy*cy);\n"
           "double gt = (G*Me)/(cr*cr);\n"
           "vx = (cx - lx) + (ax + gt*cx/cr)/2;\n"
           "vy = (cy - ly) + (ay + gt*cy/cr)/2;\n"
           "lx = cx; ly = cy;\n"
           "}\n"
           "if (output[0] != 0 && flog) {\n"
           "unsigned int tmp;\n"
           "fwrite(&timestep, sizeof(unsigned int), 1, flog);\n"
           "tmp = 0; fwrite(&tmp, sizeof(unsigned int), 1, flog);\n"
           "fclose(flog);\n"
           "flog = NULL;\n"
           "}\n"
           "return output[0];\n"
           "}\n");

    printf("static const double mem_init[] = {\n");
    for(pos = 0; pos < end; pos++) {
        printf("%.16e,\n", memory[pos]);
    }
    printf("};\n");

    printf("void set_thrust(double ax, double ay) {\n"
           "unsigned int tmp;\n"
           "if (!flog) { fprintf(stderr, \"error: already accomplished!\\n\"); exit(1); }\n"
           "ax = -ax; ay = -ay;\n"
           "input[2] = ax; input[3] = ay;\n"
           "fwrite(&timestep, sizeof(unsigned int), 1, flog);\n"
           "tmp = 2; fwrite(&tmp, sizeof(unsigned int), 1, flog);\n"
           "tmp = 2; fwrite(&tmp, sizeof(unsigned int), 1, flog);\n"
           "fwrite(&ax, sizeof(double), 1, flog);\n"
           "tmp = 3; fwrite(&tmp, sizeof(unsigned int), 1, flog);\n"
           "fwrite(&ay, sizeof(double), 1, flog);\n"
           "}\n"
           "unsigned int get_timestep() { return timestep; }\n"
           "double read_sensor(int port) { return output[port]; }\n"
           "double get_vx() { return vx; }\n"
           "double get_vy() { return vy; }\n"
           "void vm_save() {\n"
           "memcpy(save_mem, mem, sizeof(mem)); memcpy(save_input, input, sizeof(input)); memcpy(save_output, output, sizeof(output));\n"
           "save_status = status; save_flog = flog; save_timestep = timestep; save_lx = lx; save_ly = ly; save_vx = vx; save_vy = vy;\n"
           "flog = NULL; input[2] = input[3] = 0;\n"
           "}\n"
           "void vm_load() {\n"
           "memcpy(mem, save_mem, sizeof(mem)); memcpy(input, save_input, sizeof(input)); memcpy(output, save_output, sizeof(output));\n"
           "status = save_status; flog = save_flog; timestep = save_timestep; lx = save_lx; ly = save_ly; vx = save_vx; vy = save_vy;\n"
           "}\n"
           "static const unsigned int MAGIC = 0xCAFEBABE;\n"
           "static const unsigned int TEAM_ID_ = 652;\n"
           "void vm_init(unsigned int scenario_id) {\n"
           "unsigned int tmp;\n"
           "double tmpd;\n"
           "timestep = 0;\n"
           "status = 0;\n"
           "memset(mem, 0, sizeof(mem));\n"
           "memset(input, 0, sizeof(input));\n"
           "memset(output, 0, sizeof(output));\n"
           "memcpy(mem, mem_init, sizeof(mem_init));\n"
           "input[0x3e80] = (double)scenario_id;\n"
           "flog = fopen(\"trace.bin\", \"w\");\n"
           "if (!flog) { fprintf(stderr, \"failed to open trace.bin!\\n\"); exit(1); }\n"
           "fwrite(&MAGIC, sizeof(unsigned int), 1, flog);\n"
           "fwrite(&TEAM_ID_, sizeof(unsigned int), 1, flog);\n"
           "fwrite(&scenario_id, sizeof(unsigned int), 1, flog);\n"
           "tmp = 0; fwrite(&tmp, sizeof(unsigned int), 1, flog);\n"
           "tmp = 1; fwrite(&tmp, sizeof(unsigned int), 1, flog);\n"
           "tmp = 0x3e80; fwrite(&tmp, sizeof(unsigned int), 1, flog);\n"
           "tmpd = (double)scenario_id; fwrite(&tmpd, sizeof(double), 1, flog);\n"
           "}\n"
        );

    return 0;
}
