
#include <cmath>
#include <complex>
#include <vector>
#include <iostream>
#include <algorithm>
#include <ctime>
#include <cfloat>

using namespace std;
#include "../../include/ovm.hpp"
#include "moon_actions.hpp"
#include "guess_ellipse.hpp"

const P I(0., 1.);


double inner_product(const P p1, const P p2) {
  return real(conj(p1) * p2);
}

double outer_product(const P p1, const P p2) {
  return imag(conj(p1) * p2);
}

bool is_counterclockwise(const P &q, const P &p)
{
  return (outer_product(q, p) > 0.);
}

double get_direction(const P &q, const P &p)
{
  return is_counterclockwise(q, p) > 0 ? 1. : -1.;
}

double get_omega(P q)
{
  double r = abs(q);
  double w = sqrt(G * Me / (r * r * r));
  return w;
}

double get_omega(P q, P p)
{
  double r = abs(q);
  double w = sqrt(G * Me / (r * r * r)) * 
    (is_counterclockwise(q, p) ? 1. : -1.);
  return w;
}

void run_with_log(OVM &ovm)
{

  ovm.run();
  if(ovm.current_timestep() % 200 == 0)
    ovm.print_vis();
}

double time_to_transient(double r0, double r1)
{
  const double elipse_a = 0.5 * (r0 + r1);
  double time_to_change_orbits = 
    M_PI * sqrt(elipse_a * elipse_a * elipse_a /  (G*Me));
  return time_to_change_orbits;
}

void bump(OVM &ovm, double r0, double r1)
{
  const double dv = sqrt(G*Me/r0)*(sqrt((2*r1)/(r0+r1)) - 1);
  if (dv > ovm.current_fuel()) {
    // If the fuel is too much, do nothing.
    return;
  }

  P cur = ovm.current_fate_position();
  double dir0 = is_counterclockwise(cur, ovm.current_fate_velocity()) ? 1 : -1;
  P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
  cout << "#Pushed at " << thrust << " time = " << ovm.current_timestep() << endl;

  if (abs(thrust) > ovm.current_fuel()) {
    thrust = P();
  }

  ovm.print_vis();
  ovm.run(thrust);
  ovm.print_vis();
}

void arrive(OVM &ovm, double r0, double r1)
{
  const double dv = sqrt(G*Me/r1)*(1 - sqrt((2*r0)/(r0+r1)));
  if (dv > ovm.current_fuel()) {
    // If the fuel is too much, do nothing.
    return;
  }

  P cur = ovm.current_fate_position();
  // FIXME: pass direction for different direction
  
  double dir0 = is_counterclockwise(cur, ovm.current_fate_velocity()) ? 1 : -1;
  P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
  cout << "#Pushed at " << thrust << " time = " << ovm.current_timestep() << endl;
  
  if (abs(thrust) > ovm.current_fuel()) {
    thrust = P();
  }

  ovm.print_vis();
  ovm.run(thrust);
  ovm.print_vis();
}


void invert_orbital(OVM &ovm)
{
  P q = ovm.current_fate_position();
  P p = ovm.current_fate_velocity();
  double r = abs(q);
  double w = sqrt(G * Me / (r * r * r)) * - get_direction(q, p);
  // velocity = rad * r
  double target_vel = w * r;
  // optimal velocity = (~q) * I * target_vel
  P ptarg = (q / r) * I * target_vel;
  cout << "Inverting orbital. from " << p << " to " << ptarg << endl;
  P thrust = ptarg - p;
  cout << "Inverting thrust " << thrust << endl;
  if (abs(thrust) > ovm.current_fuel()) {
    cerr << "Warning: not enough fuel" << endl;
    thrust = P();
  }

  ovm.run(thrust);
}

void sync_to_circular_orbital(OVM &ovm, double r = 0., bool force_sync = false)
{
  P q = ovm.current_fate_position();
  P p = ovm.current_fate_velocity();
  if(r == 0.) r = abs(q);
  double w = sqrt(G * Me / (r * r * r)) * get_direction(q, p);
  // velocity = rad * r
  double target_vel = w * r;
  // optimal velocity = (~q) * I * target_vel
  P ptarg = (q / r) * I * target_vel;
  cout << "Adjusting velocity from " << p << " to " << ptarg << endl;
  P thrust = ptarg - p;
  cout << "Adjust thrusting " << thrust << endl;
  //  guess_ellipse_by_earth(ovm, FATE_ID);

  if (abs(thrust) > ovm.current_fuel()) {
    if (force_sync) {
      thrust *= (ovm.current_fuel() - 0.1) / abs(thrust); // 0.1 for error.
    } else {
      thrust = P();
    }
  }

  ovm.print_status();
  ovm.run(thrust);
  ovm.print_status();
  //  guess_ellipse_by_earth(ovm, FATE_ID);
}

P force_from_moon(const P &q, const P &moonq)
{
  P dx = q - moonq;
  double rd = abs(dx);
  return - G * Mm * dx / (rd * rd * rd);
}

P force_from_earth(const P &q)
{
  double r = abs(q);
  return - G * Me * q / (r * r * r);
}

struct small_simulator {
  P targetq, targetp;
  P moonq, moonp;

  P moon_earth_dv;
  P tag_earth_dv;
  P tag_moon_dv;

  P initial_thrust;

  small_simulator(OVM &ovm, int target = -1, const P &initial_thrust = P())
    : initial_thrust(initial_thrust)
  {
    if(target == -1){
      targetq = ovm.current_fate_position();
      targetp = ovm.current_fate_velocity();
    }else{
      targetq = ovm.current_target_position(target); 
      targetp = ovm.current_target_velocity(target);
    }
    moonq = ovm.current_moon_position();
    moonp = ovm.current_moon_velocity();

    recalc_force();
  }

  void recalc_force() {
    moon_earth_dv = force_from_earth(moonq);
    tag_earth_dv = force_from_earth(targetq);
    tag_moon_dv = force_from_moon(targetq, moonq);
  }
  
  void step(int steps = 1) {
    for(int i = 0; i < steps; ++i){
      targetq += 0.5 * targetp;
      moonq += 0.5 * moonp;
      recalc_force();
      targetp += tag_earth_dv + tag_moon_dv + initial_thrust;
      initial_thrust = P(0., 0.);
      moonp += moon_earth_dv;
      targetq += 0.5 * targetp;
      moonq += 0.5 * moonp;
    }
  }
};

/*
  Smart TAS-san
 */
P smarttas(OVM &original_ovm, int target, int timelimit)
{
  cout << "SMART TAS: " << original_ovm.current_fate_position() << " + " << original_ovm.current_fate_velocity() 
       << " vs " << original_ovm.current_target_position(target) << " + " << original_ovm.current_target_velocity(target) << endl;
  cout << "Initial distance: " << abs(original_ovm.current_fate_position() - original_ovm.current_target_position(target)) << endl;
  cout << "Within " << timelimit << endl;
  P thrust;
  while(true){
    if(abs(thrust) > original_ovm.current_fuel()) return thrust; // fail
    OVM trial = original_ovm;
    trial.run(thrust);

    P min_displace;
    double mincost = DBL_MAX;
    int when_min = -1;
    for(int i = 1; i < timelimit; ++i){
      P dq = trial.current_target_position(target) - trial.current_fate_position();
      double dqr = abs(dq);
      double cost = dqr;
      
      if(cost < mincost){
        min_displace = dq;
        when_min = i;
        mincost = cost;
      }
      trial.run();
    }
    cout << "Minimal displacement: " << min_displace << ", required fuel = " << abs(thrust) << endl;
    if(abs(min_displace) < 950.){
      return thrust;
    }
    thrust += min_displace / (double)when_min;
  }
}

/*
 * Yet Another TAS
 */
P yatas(OVM &original_ovm, int target, int timelimit)
{
    cout << "Yet Another TAS: " << original_ovm.current_fate_position() << " + " << original_ovm.current_fate_velocity() 
         << " vs " << original_ovm.current_target_position(target) << " + " << original_ovm.current_target_velocity(target) << endl;
    cout << "Initial distance: " << abs(original_ovm.current_fate_position() - original_ovm.current_target_position(target)) << endl;
    cout << "Within " << timelimit << endl;

    // まず、最接近点を求める
    int sample_time = -1;
    P sample_original_pos;
    P nearest_diff;
    {
        double sample_distance = 1e100;
        OVM trial(original_ovm);
        trial.run();
        for(int i = 1; i < timelimit; i++) {
            P diff = trial.current_target_position(target) - trial.current_fate_position();
            if (abs(diff) < sample_distance) {
                sample_time = i;
                sample_original_pos = trial.current_fate_position();
                sample_distance = abs(diff);
                nearest_diff = diff;
            }
            trial.run();
        }
    }

    // (1, 0) をthrustして実行してみる
    P delta;
    {
        OVM trial(original_ovm);
        trial.run(P(1, 0));
        for(int i = 1; i < sample_time; i++)
            trial.run();
        delta = (trial.current_fate_position() - sample_original_pos) / (double)sample_time;
    }

    P thrust = nearest_diff / (delta * (double)sample_time);
    while(true){
        if(abs(thrust) > original_ovm.current_fuel()) return thrust; // fail
        OVM trial = original_ovm;
        trial.run(thrust);

        P min_displace;
        double mincost = DBL_MAX;
        int when_min = -1;
        for(int i = 1; i < timelimit; ++i){
            P dq = trial.current_target_position(target) - trial.current_fate_position();
            double dqr = abs(dq);
            double cost = dqr;
      
            if(cost < mincost){
                min_displace = dq;
                when_min = i;
                mincost = cost;
            }
            trial.run();
        }
        cout << "Minimal displacement: " << min_displace << ", required fuel = " << abs(thrust) << endl;
        if(abs(min_displace) < 950.){
            return thrust;
        }
        thrust += min_displace / (delta * (double)when_min);
    }
}

/*
  TAS-san
*/
P tasmachine(OVM &original_ovm, int target, int timelimit)
{
  // estimate target position
  OVM tracer = original_ovm;
  P min_displace;
  double mincost = DBL_MAX;
  int when_min = -1;
  for(int i = 1; i < timelimit; ++i){
    tracer.run();
    P dq = tracer.current_target_position(target) - tracer.current_fate_position();
    double dqr = abs(dq);
    double fuelcost = dqr / i;
    
    if(fuelcost < mincost){
      min_displace = dq;
      when_min = i;
      mincost = fuelcost;
    }
  }

  assert(when_min != -1);
  double scale = mincost / 20;
  P thrust = min_displace / (double) when_min;
  for(int cnt = 0; cnt < 1000; ++cnt){
    double radius = 950.; // < 1km, for reducing error
    double r1 = rand() / (RAND_MAX + 1.0); double r1fac = scale * sqrt(-2. * log(r1));
    double r2 = rand() / (RAND_MAX + 1.0); double r2fac = 2 * M_PI * r2;
    P thrust = min_displace / (double)when_min + P(r1fac * cos(r2fac), r1fac * sin(r2fac));// Box-Muller method
    OVM trial = original_ovm;
    trial.run(thrust);

    for(int i = 0; i < timelimit; ++i){
      trial.run();
      if(abs(trial.current_target_position(target) - trial.current_fate_position()) < radius)
        return thrust;
    }
  }
  return P(1e+10, 0.);
}

/*
  pass-by target entity (0-11 == debris, 12 == moon)
 */
bool propose_sync_to(int entity, int time, OVM &ovm, OVM& myovm, double &whenbump, double &whenarrive, double &r1, double &fuel)
{
  int dt = time - ovm.current_timestep();
  if(dt < 0)
    return false;

  //  OVM myovm = ovm;
  while(myovm.current_timestep() < time)
    myovm.run();
  P p, q;
  if(entity == 12){
    // moon
    q = myovm.current_moon_position();
    p = myovm.current_moon_velocity();
  }else{
    q = myovm.current_target_position(entity);
    p = myovm.current_target_velocity(entity);

#if 0
    // moon gravity correction (linear interpolation)
    P moon = myovm.current_moon_position();
    double rmoon = abs(moon);
    P gmoon = G * Mm * moon / (rmoon * rmoon * rmoon);

    q -= 0.5 * gmoon * (double)(dt * dt);
    p -= gmoon * (double)dt;
#endif
  }
  
  P start = ovm.current_fate_position();
  P startvel = ovm.current_fate_velocity();
  
  double r0 = abs(start);
  r1 = abs(q);

  double w0 = get_omega(start, startvel);
  double w1 = get_omega(q) * (w0 > 0 ? 1 : -1); // w1 should have equal sign to w0
  //cout << "w0 = " << w0 << " w1 = " << w1 << endl;
  
  double theta0 = arg(start);
  double theta1 = arg(q);
  double delta_theta = theta1 - theta0;

  const double elipse_a = 0.5 * (r0 + r1);
  double time_to_change_orbits = 
    M_PI * sqrt(elipse_a * elipse_a * elipse_a /  (G*Me));
  double tau = dt - time_to_change_orbits;
  //  double t1minimum = max(400, dt / 30); //510.; // max thrust ~= 
  double t1minimum = 510.;

  if (tau < t1minimum) {
    return false;
  }

  // t0 = tau - t1_minimum, t1 = t1_miniumu
  double theta_bound0 = w0 * (tau - t1minimum) + w1 * t1minimum + M_PI - delta_theta;
  // t0 = 0, t1 = tau
  double theta_bound1 = w1 * tau + M_PI - delta_theta;
  
  int phase0 = (int)floor(theta_bound0 / (M_PI * 2));
  int phase1 = (int)floor(theta_bound1 / (M_PI * 2));
  if(phase0 == phase1){
    // no integer solution
    return false;
  }
  double n;
  if(phase0 < phase1)
    n = phase0 + 1.0;
  else
    n = phase1 + 1.0;
  
  double t0, t1;
  t0 = (M_PI * 2 * n - tau * w1 - M_PI + delta_theta) / (w0 - w1);
  t1 = tau - t0;

  double dv0 = sqrt(G*Me/r0)*(sqrt((2*r1)/(r0+r1)) - 1);
  double dv1 = sqrt(G*Me/r1)*(1 - sqrt((2*r0)/(r0+r1)));
  fuel = fabs(dv0) + fabs(dv1);
  // fuel = dv0 + dv1;
  if (fuel > ovm.current_fuel()) {
    return false;
  }

  whenbump = ovm.current_timestep() + t0; 
  whenarrive = time - t1;
  /*
  cout << "t0 " << t0 << " t1 " << t1 << " tau " << tau << endl;
  cout << "time-to-change " << time_to_change_orbits << endl;
  cout << "Whenbump: " << whenbump << endl;
  cout << "Whenarrive: " << whenarrive << " should be: "
       << ovm.current_timestep() + t0 + time_to_change_orbits << endl;
  cout << "Estmated fuel: " << fuel << endl;
  */
  return true;
}

void process_transient(OVM &ovm, int entity, int target_time, double whenbump, double whenarrive, double r1)
{
  // ganbaru
  cout << "Target entity = " << entity << ", time = " << ovm.current_timestep() << " target_time: " << target_time << endl;
  while(ovm.current_timestep() + .5 < whenbump) run_with_log(ovm);
  double r0 = abs(ovm.current_fate_position());
  cout << "bump! r0 = " << r0 << " r1 = " << r1 << endl;
  bump(ovm, r0, r1);
  P thrust;
  while(ovm.current_timestep() + .5 < whenarrive) run_with_log(ovm);
  arrive(ovm, r0, r1);
  cout << "Lanuch finished, current height = " << abs(ovm.current_fate_position()) << " r1 = " << r1 << endl;
  int super_tas_time = 500;
  int sync_loop_len =
    max(500, target_time - ovm.current_timestep() / 100);
  int loop_counter = sync_loop_len;
  while(ovm.current_timestep() <= target_time - super_tas_time){
    /*
    // sync to orbit, as possible
    if (--loop_counter == 0) {
      //if(ovm.current_timestep() % 100 == 0){
      sync_to_circular_orbital(ovm, r1);
      loop_counter = sync_loop_len;
      cout << "Resync to orbit, current height = " << abs(ovm.current_fate_position()) << " r1 = " << r1 << " fuel: " << ovm.current_fuel() << endl;
    }
    */
    run_with_log(ovm);
  }
  // Super tas time!
  //thrust = tasmachine(ovm, entity, target_time + super_tas_time - ovm.current_timestep());
  thrust = smarttas(ovm, entity, target_time + super_tas_time - ovm.current_timestep());
  cout << "TAS thrust " << thrust << endl;
  if(abs(thrust) < ovm.current_fuel()) {
    ovm.print_vis();
    ovm.run(thrust);
    ovm.print_vis();
    while(!ovm.target_collected(entity))
      run_with_log(ovm);
  }else{
    cerr << "Warning: too much thrust, requested = " << thrust << ", but tank = " << ovm.current_fuel() << endl;
  }
  double dist = abs(ovm.current_target_position(entity)
                    - ovm.current_fate_position());
  cout << "Distance target: " << ovm.current_fate_position() << " - "
       << ovm.current_target_position(entity) << ": "
       << dist
       << " Current vel: " << ovm.current_fate_velocity()
       << ", dist / vel = " << dist / abs(ovm.current_fate_velocity())
       << endl;
  sync_to_circular_orbital(ovm);
}

// quick hack
const int MOON_ID = -3;

void hohmann_params_to_entity(OVM &ovm, int entity_id, double &tbump, double &tarrive)
{
  sync_to_circular_orbital(ovm);
  P fate = ovm.current_fate_position();
  P target = (entity_id == MOON_ID ? ovm.current_moon_position() : ovm.current_target_position(entity_id));
  P targetv = (entity_id == MOON_ID ? ovm.current_moon_velocity() : ovm.current_target_velocity(entity_id));
  double cur_r = abs(fate);
  double t_orbit = abs(target);

  double theta0 = arg(fate);
  double theta1 = arg(target);
  double w0 = get_omega(fate, ovm.current_fate_velocity());
  double w1 = get_omega(target, ovm.current_target_velocity(entity_id));
  double ttime = time_to_transient(cur_r, t_orbit);

  // theta0 + T w0 + pi = theta1 + (T + ttime) w1 +- 2 pi n
  // T(w0 - w1) = theta1 - theta0 - pi + ttime w1 +- 2 pi n
  //            == dtheta + 2 pi n
  // T = (dtheta +- 2 pi n) / (w0 - w1)
  double dtheta = theta1 - theta0 - M_PI + ttime * w1;
  if(w0 - w1 > 0.){
    while(dtheta < 0.) dtheta += 2 * M_PI; // ceil? nani sore?
  }else{
    while(dtheta > 0.) dtheta -= 2 * M_PI;
  }
  tbump = dtheta / (w0 - w1);
  assert(tbump > 0.);
  tarrive = tbump + ttime;
}

void refuel_to_station(OVM &ovm)
{
  cout << "Remaining fuel: " << ovm.current_fuel() << endl;

  double tbump, tarrive;
  sync_to_circular_orbital(ovm);
  hohmann_params_to_entity(ovm, STATION_ID, tbump, tarrive);
  P fate = ovm.current_fate_position();
  P station = ovm.current_station_position();
  double cur_r = abs(fate);
  double loworbit = abs(station);

  for(int i = 0; i < tbump; ++i){ run_with_log(ovm); }
  bump(ovm, cur_r, loworbit);
  for(int i = 0; i < (tarrive - tbump) - 100; ++i){ run_with_log(ovm); }
  
  cout << "Before-yatas fuel: " << ovm.current_fuel() << endl;
  P thrust = yatas(ovm, STATION_ID, 200);
  ovm.run(thrust);

  for(int i = 0; i < 200; ++i){
    run_with_log(ovm);
  }

  cout << "After-sync fuel: " << ovm.current_fuel() << endl;
}

void refuel(OVM &ovm){
  refuel_to_station(ovm);
  while(true){
    // return from my yasara-distorted orbital
    ellipse_t epl = guess_ellipse_by_earth(ovm);
    if(epl.next_apogee_time < epl.next_perigee_time || abs(epl.perigee) > Re){
      // wait for apogee
      cout << "ellip now: " << ovm.current_timestep() << " nextap: " << epl.next_apogee_time << endl;
      while(ovm.current_timestep() < epl.next_apogee_time){ run_with_log(ovm); }
      sync_to_circular_orbital(ovm);
      break;
    }else{
      // warning: NOW you are dropping. ZAKUs don't have the equipment!!
      double h = abs(ovm.current_fate_position());
      cerr << "ZAKU-mode, time = " << ovm.current_timestep() << ", h = " << h << endl;
      P thrust = ovm.current_fate_position() / h * I *
        get_direction(ovm.current_fate_position(), ovm.current_fate_velocity()) *
        10000.0 / (h - Re);
      ovm.run(thrust);
      continue;
    }
  }
}

int main_random(OVM& ovm);

int main(int argc, char* argv[]) {
  if (argc <= 1) {
    cerr << "Usage: " << argv[0] << " [senario]" << endl;
    return 1;
  }

  OVM ovm(atoi(argv[1]));
  ovm.set_auto_save(true);
  ovm.set_anti_moon(true);

  int rev = (argc <= 2 ? 0 : atoi(argv[2]));
  if (rev == -1)
    return main_random(ovm);

  // At the first run two steps to get the positions and the speed.
  ovm.run();
  double maxfuel = ovm.current_fuel();
  ovm.run();

  vector<int> candidates;
  vector<int> moon_candidates;
  for(int i = 0; i < MAX_TARGETS; ++i) {
    double rearth2 = norm(ovm.current_fate_position());
    double rmoon2 = norm(target_position_moon(ovm, i));
    if(Me / rearth2 > Mm / rmoon2){
      candidates.push_back(i);
      cerr << "candidates: " << i << endl;
    }else{
      moon_candidates.push_back(i);
      cerr << "moon: " << i << endl;
    }
  }

  reverse(candidates.begin(), candidates.begin()+rev);
  //  candidates.push_back(0);
  time_t seed = time(NULL);
  cout << "seed: " << seed << endl;
  srand(seed);
  //random_shuffle(candidates.begin(), candidates.end());
  
  for(size_t i = 0; i < candidates.size(); ++i){
    cerr << "entity: " << candidates[i] << endl;
    for (int i1 = 0; i1 < MAX_TARGETS; ++i1) {
      cerr << "#Entity: " << i1 << (ovm.target_collected(i1) ? " collected" : " not collected.") << endl;
    }

    int entity = candidates[i];
    if(ovm.target_collected(entity))
      continue;
    // TODO: atode modotte kuru
    //    vector<int> time_list;
    // for(int j = 0; j < 100000; ++j){
    //  int target_time = curtime + (int)(rand() / (RAND_MAX + 1.0) * 100000);
    //  time_list.push_back(target_time);
    // }
    //sort(time_list.begin(), time_list.end());

    if(ovm.current_fuel() < 0.3 * maxfuel){
      cerr << "Refueling" << endl;
      refuel(ovm);
    }

    cerr << "Current Fuel: " << ovm.current_fuel() << endl;
    int curtime = ovm.current_timestep();

    OVM myovm(ovm);
    double maxscore = -DBL_MAX;
    double min_whenbump, min_whenarrive, min_r1, min_fuel = DBL_MAX;
    int target_time = -1;
    for (int j = curtime; j < 2000000; ++j) {
      if (j % 10000 == 0) {
        cerr << "j: " << j << endl;
      }
      double whenbump, whenarrive, r1, fuel;
      if (propose_sync_to(entity, j, ovm, myovm, whenbump, whenarrive, r1, fuel)) {
	double score = 75 * (2e6 - j) * (candidates.size() - i) / 24e6 - 25 * fuel / maxfuel;
	//cerr << score << endl;
	//	cerr << maxscore << ", " << score << endl;
	if (score > maxscore) {
	  maxscore = score;
          min_whenbump = whenbump;
          min_whenarrive = whenarrive;
          min_r1 = r1;
          min_fuel = fuel;
          target_time = j;
	  //	  cerr << "target time: " << target_time << endl;
        }
	//	break;
      }
    }
    if (target_time < 0) {
      cerr << "Failed to find." << endl;
      continue;
    }
    cerr << "Found: " << target_time << ", " << min_fuel << endl;
    cerr << "score: " << maxscore << endl;

    double
      whenbump = min_whenbump,
      whenarrive = min_whenarrive,
      r1 = min_r1;
    process_transient(ovm, entity, target_time, whenbump, whenarrive, r1);

    if (rev > 0 && i == 0)
        ovm.run(-2.0*ovm.current_fate_velocity());
  }

  // With current scoring rules, better scores are obtained with ignoring moon
#if 0
  // refuel to get to the moon!
  refuel_to_station(ovm);
  sync_to_circular_orbital(ovm);

  {
    double tbump, tarrive;
    hohmann_params_to_entity(ovm, MOON_ID, tbump, tarrive);
    for(int i = 0; i < tbump; ++i){ run_with_log(ovm); }
    ovm.set_anti_moon(false);
    // bump!
    double r0 = abs(ovm.current_fate_position());
    double r1 = abs(ovm.current_moon_position());
    bump(ovm, r0, r1);
    // stop at 4000 km
    while(abs(ovm.current_fate_position() - ovm.current_moon_position()) > 4e6){
      run_with_log(ovm);
    }
    // thrust to get moon's orbital
    sync_to_circular_orbital_moon(ovm);
    // search entity & catch!
    for(size_t i = 0; i < moon_candidates.size(); ++i){
      int target = moon_candidates[i];
      sync_to_circular_orbital_moon(ovm);
      P targq = target_position_moon(ovm, target);
      P targp = target_velocity_moon(ovm, target);
      P fateq = fate_position_moon(ovm);
      P fatep = fate_velocity_moon(ovm);
      double r0 = abs(fateq);
      double r1 = abs(targq);
      cerr << "r0: " << r0 << " r1: " << r1 << endl;

      double tbump, tarrive;
      hohmann_params_to_entity_moon(ovm, target, tbump, tarrive);

      for(int i = 0; i < tbump; ++i){
        run_with_log(ovm);
      }
      bump_moon(ovm, r0, r1);
      for(int i = 0; i < (tarrive-tbump) - 100; ++i){
        run_with_log(ovm);
      }
      arrive_moon(ovm, r0, r1);
      P thrust = yatas(ovm, target, 200);
      ovm.run(thrust);
      for(int i = 0; i < 200; ++i){
        run_with_log(ovm);
      }
      for (int i1 = 0; i1 < MAX_TARGETS; ++i1) {
        cerr << "#Entity: " << i1 << (ovm.target_collected(i1) ? " collected" : " not collected.") << endl;
      }
    }
  }
#endif

  cerr << "Run to the limit-time." << endl;
  //  sync_to_circular_orbital(ovm, 0, true);
  while (ovm.run_free()) {
  }

  cout << "Score: " << ovm.current_score() << endl;
  cout << "Time: " << ovm.current_timestep() << endl;
  return 0;
}

int main_random(OVM& ovm) {
  // At the first run two steps to get the positions and the speed.
  ovm.run();
  double maxfuel = ovm.current_fuel();
  ovm.run();

  vector<int> candidates;
  vector<int> moon_candidates;
  for(int i = 0; i < MAX_TARGETS; ++i) {
    double rearth2 = norm(ovm.current_fate_position());
    double rmoon2 = norm(target_position_moon(ovm, i));
    if(Me / rearth2 > Mm / rmoon2){
      candidates.push_back(i);
    }else{
      moon_candidates.push_back(i);
    }
  }
  //  candidates.push_back(0);
  time_t seed = time(NULL);
  cout << "seed: " << seed << endl;
  srand(seed);
  //random_shuffle(candidates.begin(), candidates.end());
  
  for(size_t i = 0; i < candidates.size(); ++i){
    for (int i1 = 0; i1 < MAX_TARGETS; ++i1) {
      cerr << "#Entity: " << i1 << (ovm.target_collected(i1) ? " collected" : " not collected.") << endl;
    }

    // TODO: atode modotte kuru
    //    vector<int> time_list;
    // for(int j = 0; j < 100000; ++j){
    //  int target_time = curtime + (int)(rand() / (RAND_MAX + 1.0) * 100000);
    //  time_list.push_back(target_time);
    // }
    //sort(time_list.begin(), time_list.end());

    if(ovm.current_fuel() < 0.3 * maxfuel){
      cerr << "Refueling" << endl;
      refuel(ovm);
    }

    cerr << "Current Fuel: " << ovm.current_fuel() << endl;
    int curtime = ovm.current_timestep();

    OVM myovm(ovm);
    double maxscore = -DBL_MAX;
    double min_whenbump, min_whenarrive, min_r1, min_fuel = DBL_MAX;
    int target_time = -1;
    int target_entity = -1;
    double each_max[MAX_TARGETS];
    int each_time[MAX_TARGETS];
    for (int k = 0; k < candidates.size(); ++k) {
      const int entity = candidates[k];
      each_max[entity] = -DBL_MAX;
      each_time[entity] = 0;
    }
    for (int j = curtime; j < 2000000; ++j) {
      if (j % 10000 == 0) {
        cerr << "j: " << j << endl;
      }
      for (int k = 0; k < candidates.size(); ++k) {
	const int entity = candidates[k];
	if(ovm.target_collected(entity))
	  continue;
	double whenbump, whenarrive, r1, fuel;
	if (propose_sync_to(entity, j, ovm, myovm,
			    whenbump, whenarrive, r1, fuel)) {
	  double score = 75 * (2e6 - j) * (candidates.size() - i) / 24e6
	    - 25 * fuel / maxfuel;
	  //cerr << score << endl;
	  //    cerr << maxscore << ", " << score << endl;
	  if (score > maxscore) {
	    maxscore = score;
	    min_whenbump = whenbump;
	    min_whenarrive = whenarrive;
	    min_r1 = r1;
	    min_fuel = fuel;
	    target_time = j;
	    target_entity = entity;
	    //    cerr << "target time: " << target_time << endl;
	  }
	  if (score > each_max[entity]) {
	    each_max[entity] = score;
	    each_time[entity] = j;
	  }
	}
      }
    }
    for (int k = 0; k < candidates.size(); ++k) {
      const int entity = candidates[k];
      if (!ovm.target_collected(entity)) {
	cerr << "Entity#" << entity << " max: " << each_max[entity]
	     << " " << each_time[entity] << endl;
	if (target_time < 0) {
	  cerr << "Failed to find." << endl;
	  continue;
	}
      }
    }
    cerr << "Found: " << target_time << ", " << min_fuel << endl;
    cerr << "score: " << maxscore << endl;

    double
      whenbump = min_whenbump,
      whenarrive = min_whenarrive,
      r1 = min_r1;
    // Calc fuel for r1 -> station hohman jump.
    {
      double r0 = abs(ovm.current_station_position());
      double req_fuel = fabs(sqrt(G*Me/r1)*(sqrt((2*r0)/(r0+r1)) - 1));
      //+ fabs(sqrt(G*Me/r0)*(1 - sqrt((2*r1)/(r0+r1))));
      cerr << "min fuel, req_fuel, current_fuel:"
          << min_fuel << " "
          << req_fuel << " "
          << ovm.current_fuel() << endl;
      if (ovm.current_fuel() < req_fuel + min_fuel + 2000) {
       cerr << "Cancel and refueling" << endl;
       refuel(ovm);
       --i;
       continue;
      }
    }
    process_transient(ovm, target_entity, target_time, whenbump, whenarrive, r1);
  }

  cerr << "Run to the limit-time." << endl;
  //  sync_to_circular_orbital(ovm, 0, true);
  while (ovm.run_free()) {
  }

  cout << "Score: " << ovm.current_score() << endl;
  cout << "Time: " << ovm.current_timestep() << endl;
  return 0;
}
