#include <cmath>
#include "moon_actions.hpp"

static const P I(0., 1.);

static double outer_product(const P p1, const P p2) {
  return imag(conj(p1) * p2);
}

static bool is_counterclockwise(const P &q, const P &p)
{
  return (outer_product(q, p) > 0.);
}

static double get_direction(const P &q, const P &p)
{
  return is_counterclockwise(q, p) > 0 ? 1. : -1.;
}

static double get_omega(P q, P p)
{
  double r = abs(q);
  double w = sqrt(G * Mm / (r * r * r)) * 
    (is_counterclockwise(q, p) ? 1. : -1.);
  return w;
}

static double time_to_transient(double r0, double r1)
{
  const double elipse_a = 0.5 * (r0 + r1);
  double time_to_change_orbits = 
    M_PI * sqrt(elipse_a * elipse_a * elipse_a /  (G*Mm));
  return time_to_change_orbits;
}

P fate_position_moon(OVM &ovm)
{
  return ovm.current_fate_position() - ovm.current_moon_position();
}

P fate_velocity_moon(OVM &ovm)
{
  return ovm.current_fate_velocity() - ovm.current_moon_velocity();
}

P target_position_moon(OVM &ovm, int id)
{
  return ovm.current_target_position(id) - ovm.current_moon_position();
}

P target_velocity_moon(OVM &ovm, int id)
{
  return ovm.current_target_velocity(id) - ovm.current_moon_velocity();
}

void bump_moon(OVM &ovm, double r0, double r1)
{
  const double dv = sqrt(G*Mm/r0)*(sqrt((2*r1)/(r0+r1)) - 1);
  if (dv > ovm.current_fuel()) {
    // If the fuel is too much, do nothing.
    return;
  }

  P cur = fate_position_moon(ovm);
  double dir0 = is_counterclockwise(cur, fate_velocity_moon(ovm)) ? 1 : -1;
  P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;

  if (abs(thrust) > ovm.current_fuel()) {
    thrust = P();
  }

  ovm.print_vis();
  ovm.run(thrust);
  ovm.print_vis();
}

void arrive_moon(OVM &ovm, double r0, double r1)
{
  const double dv = sqrt(G*Mm/r1)*(1 - sqrt((2*r0)/(r0+r1)));
  if (dv > ovm.current_fuel()) {
    // If the fuel is too much, do nothing.
    return;
  }

  P cur = fate_position_moon(ovm);
  // FIXME: pass direction for different direction
  
  double dir0 = is_counterclockwise(cur, fate_velocity_moon(ovm)) ? 1 : -1;
  P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
  
  if (abs(thrust) > ovm.current_fuel()) {
    thrust = P();
  }

  ovm.print_vis();
  ovm.run(thrust);
  ovm.print_vis();
}

void sync_to_circular_orbital_moon(OVM &ovm)
{
  P q = ovm.current_fate_position() - ovm.current_moon_position();
  P p = ovm.current_fate_velocity() - ovm.current_moon_velocity();
  double r = abs(q);
  double w = sqrt(G * Mm / (r * r * r)) * get_direction(q, p);
  // velocity = rad * r
  double target_vel = w * r;
  // optimal velocity = (~q) * I * target_vel
  P ptarg = (q / r) * I * target_vel;
  //cout << "Adjusting velocity from " << p << " to " << ptarg << endl;
  P thrust = ptarg - p;
  //cout << "Adjust thrusting " << thrust << endl;
  if (abs(thrust) > ovm.current_fuel()) {
    thrust = P();
  }

  ovm.run(thrust);
}


void hohmann_params_to_entity_moon(OVM &ovm, int entity_id, double &tbump, double &tarrive)
{
  sync_to_circular_orbital_moon(ovm);
  P fate = fate_position_moon(ovm);
  P target = target_position_moon(ovm, entity_id);
  P targetv = target_velocity_moon(ovm, entity_id);
  double cur_r = abs(fate);
  double t_orbit = abs(target);

  double theta0 = arg(fate);
  double theta1 = arg(target);
  double w0 = get_omega(fate, fate_velocity_moon(ovm));
  double w1 = get_omega(target, targetv);
  double ttime = time_to_transient(cur_r, t_orbit);

  // theta0 + T w0 + pi = theta1 + (T + ttime) w1 +- 2 pi n
  // T(w0 - w1) = theta1 - theta0 - pi + ttime w1 +- 2 pi n
  //            == dtheta + 2 pi n
  // T = (dtheta +- 2 pi n) / (w0 - w1)
  double dtheta = theta1 - theta0 - M_PI + ttime * w1;
  if(w0 - w1 > 0.){
    while(dtheta < 0.) dtheta += 2 * M_PI; // ceil? nani sore?
  }else{
    while(dtheta > 0.) dtheta -= 2 * M_PI;
  }
  tbump = dtheta / (w0 - w1);
  assert(tbump > 0.);
  tarrive = tbump + ttime;
}
