#include <algorithm>
#include <complex>
#include <iomanip>
#include <iostream>
#include <cassert>
#include <cmath>
#include <cstring>
#include "ovm.hpp"

using namespace std;

#include "transitions.hpp"
#include "guess_ellipse.hpp"

const P I(0, 1);

bool wait_perigee(OVM& ovm) {
    ellipse_t e = guess_ellipse_by_earth(ovm);

    if (e.perigee == P()) {
        printf("wait_perigee: perigee is too far!\n");
        return false;
    }

    printf("wait_perigee: wait %.0f seconds\n", e.next_perigee_time - ovm.current_timestep());
    while(e.next_perigee_time - ovm.current_timestep() > 0.5)
        ovm.run();
    return true;
}

void enter_circular_orbit(OVM& ovm) {
    double dir = (arg(ovm.current_fate_position() / ovm.previous_fate_position()) > 0 ? 1 : -1);

    double start_fuel = ovm.current_fuel();
    for(int i = 0; i < 10; i++) {
        P cur = ovm.current_fate_position();
        double r = abs(cur);
        P expected_velocity = (cur / r) * I * dir * sqrt(G*Me/r);
        P thrust = expected_velocity - ovm.current_fate_velocity();
        ovm.run(thrust);
    }
    double end_fuel = ovm.current_fuel();
    printf("enter_circular_orbit: entered to orbit = %.0f, took %.0f fuel\n", abs(ovm.current_fate_position()), start_fuel-end_fuel);
}


bool transit_circular_orbit(OVM& ovm, P target_point, int target_timestep) {
    double dir = (arg(ovm.current_fate_position() / ovm.previous_fate_position()) > 0 ? 1 : -1);
    double target_orbital = abs(target_point);
    const P start = ovm.current_fate_position();
    const double r0 = abs(start);
    const double r1 = target_orbital;

    printf("transit_circular_orbit: requested to transit from %.0f to %.0f, arriving point (%.0f, %.0f) at time %d\n",
           r0, r1, target_point.real(), target_point.imag(), target_timestep);

    double start_fuel = ovm.current_fuel();

    // Time required to move from start orbit to the goal orbit.
    const double elipse_a = 0.5 * (r0 + r1);
    const double tt_change_orbits = M_PI *
        sqrt(elipse_a * elipse_a * elipse_a /  (G*Me));
    const double w0 = sqrt(G * Me / (r0 * r0 * r0)) * dir;
    const double w1 = sqrt(G * Me / (r1 * r1 * r1)) * dir;
  
    // Solve this.
    // theta: theta0 + w0 t0 + w1 t1 + pi = theta1 + 2 pi n
    // time:  t0 + t1 + T = Tperigee - Tcur + Telliptic k
    // where
    //        T = tt_change_orbits.
    //        Tperigee = target_time.
    //        Tcur = now.
    //        Telliptic = target_cycle.
    //
    // k is a non-negative integer, and n is an integer.
    double theta1 = arg(target_point);
    double theta0 = arg(start);

    double remain_time = target_timestep - ovm.current_timestep() - tt_change_orbits;
    if(remain_time < 0) {
        printf("transit_circular_orbit:    impossible.\n");
        return false;
    }

    // now t0 + t1 = remain_time
    double dtheta = theta1 - theta0;
    
    double bound0 = w1 * remain_time - dtheta + M_PI;
    double phase0 = floor(bound0 / (M_PI * 2));
    double bound1 = w0 * remain_time - dtheta + M_PI;
    double phase1 = floor(bound1 / (M_PI * 2));

    if((int)phase0 == (int)phase1){
        printf("transit_circular_orbit:    impossible.\n");
        return false;
    }

    double n = min(phase0, phase1) + 1.0;
    double v = dtheta + n * (M_PI * 2) - M_PI;  // v = w0 t0 + w1 t1.
    // Here, w0 t0 + w1 t1 = v.
    //       t0 + t1 = remain_time.
    // Solve this.
    double t0 = (v - w1 * remain_time) / (w0 - w1);
    double t1 = remain_time - t0;

    assert(t0 > 0);
    assert(t1 > 0);

    double start_push = ovm.current_timestep() + t0;
    double again_push = start_push + tt_change_orbits;
    double final_move = again_push + t1;

    while(start_push - ovm.current_timestep() > 0.5) {
        ovm.run();
    }

    {
        // Start push.
        double r0 = abs(ovm.current_fate_position());
        double r1 = abs(target_point);
        {
            const double dv = sqrt(G*Me/r0)*(sqrt((2*r1)/(r0+r1)) - 1);
            P cur = ovm.current_fate_position();
            P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir;

            ovm.run(thrust);
        }

        while (again_push - ovm.current_timestep() > 0.5) {
            ovm.run();
        }

        // Again push, to climb up (or down) to second circular orbital
        {
            const double dv = sqrt(G*Me/r1)*(1 - sqrt((2*r0)/(r0+r1)));

            P cur = ovm.current_fate_position();
            P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir;

            ovm.run(thrust);
        }
    }

    OVM ovm_predict(ovm);
    while(final_move - ovm_predict.current_timestep() > 0.5)
        ovm_predict.run();

    double end_fuel = ovm.current_fuel();
    printf("transit_circular_orbit: successfully transitted! now time %d, simluated to meet at time %d point (%.0f, %.0f), took %.0f fuel\n",
           ovm.current_timestep(), ovm_predict.current_timestep(),
           ovm_predict.current_fate_position().real(), ovm_predict.current_fate_position().imag(),
           start_fuel - end_fuel);

    return true;
}
