#include <algorithm>
#include <complex>
#include <iomanip>
#include <iostream>
#include <cassert>
#include <cmath>
#include <cstring>
#include "ovm.hpp"
#include "../clearsky/autofollow.hpp"

#define LOG_VIS 1

using namespace std;

const P I(0, 1);

const double PI=3.141592653589793238462;

double inner_product(const P p1, const P p2) {
  return real(conj(p1) * p2);
}

double outer_product(const P p1, const P p2) {
  return imag(conj(p1) * p2);
}

// returns (d_theta + 2 * n * pi / dw) which is larger than least_ans.
double theta_div(double d_theta, double dw, double least_ans) {
  if (dw < 0) {
    d_theta = -d_theta;
    dw = -dw;
  }
  const double least_theta = dw * least_ans;
  if (d_theta > least_ans + 2 * PI) {
    double c = floor((d_theta - least_theta) / (2 * PI));
    d_theta -= c * 2 * PI;
  }
  if (d_theta < least_theta) {
    double c = ceil((least_theta - d_theta) / (2 * PI));
    d_theta += c * 2 * PI;
  }
  return d_theta / dw;
}

/*
  Given point q and velocity v, return the shape of ellipse,
  with long_r and T.
 */
void getellipse(bool myself,
		OVM& ovm_orig,
                double &long_r, double &T,
                P &anotherfocus,
                P &apogee, P &perigee,
                double &time_apogee, double &time_perigee)
{
  OVM ovm(ovm_orig);
  time_apogee = -1;
  time_perigee = -1;
  for(int i = 0; i < 3000000; ++i){
    int now = ovm.current_timestep();
    P cur = myself ? ovm.current_fate_position()
      : ovm.current_target_position();
    P prev = myself ? ovm.previous_fate_position()
      : ovm.previous_target_position();
    ovm.run();
    P next = myself ? ovm.current_fate_position() : ovm.current_target_position();
    // double qr = abs(qsim);
    if (norm(prev) < norm(cur) &&
	norm(next) < norm(cur)) {
      apogee = cur;
      if(time_apogee != -1){
        T = now - time_apogee;
        break;
      }
      time_apogee = now;
    }
    if(norm(cur) < norm(prev) &&
       norm(cur) < norm(next)) {
      perigee = cur;
      if (time_perigee == -1) {
	time_perigee = now;
      }
    }
  }
  long_r = abs(apogee - perigee) * 0.5;
  anotherfocus = apogee + perigee;
}

double period_of_ellipse(double a, double m) {
  return 2 * PI * sqrt(a * a * a /  (G*m));
}

// To reach target_point at target_time + k * target_cycle.
void future_estimation(OVM& ovm, double dir0, double dir1,
                       const P& target_point,
                       double target_time,
                       double target_cycle,
                       double& start_push, double& again_push, 
                       double& final_move) {
  cout << "Target: " << target_point << " " << abs(target_point) << " at " << target_time << endl;
  double target_orbital = abs(target_point);
  const P start = ovm.current_fate_position();
  //const P target = current_target();
  const double r0 = abs(start);
  const double r1 = target_orbital;

  // Time required to move from start orbit to the goal orbit.
  const double tt_change_orbits = 0.5 * period_of_ellipse(0.5 * (r0 + r1), Me);
  const double w0 = sqrt(G * Me / (r0 * r0 * r0)) * dir0;
  const double w1 = sqrt(G * Me / (r1 * r1 * r1)) * dir1;
  
  // Solve this.
  // theta: theta0 + w0 t0 + w1 t1 + pi = theta1 + 2 pi n
  // time:  t0 + t1 + T = Tperigee - Tcur + Telliptic k
  // where
  //        T = tt_change_orbits.
  //        Tperigee = target_time.
  //        Tcur = now.
  //        Telliptic = target_cycle.
  //
  // k is a non-negative integer, and n is an integer.
  double theta1 = arg(target_point);
  double theta0 = arg(start);
  double t0, t1;
  for(int k = 0; /* */; ++k) {
    double real_target_time = target_time + k * target_cycle;
    double remain_time = real_target_time -
      ovm.current_timestep() - tt_change_orbits;
    if(remain_time < 0) continue;
    cout << "remain " << remain_time << endl;
    // now t0 + t1 = remain_time
    double dtheta = theta1 - theta0;
    
    double bound0 = w1 * remain_time - dtheta + M_PI;
    double phase0 = floor(bound0 / (M_PI * 2));
    double bound1 = w0 * remain_time - dtheta + M_PI;
    double phase1 = floor(bound1 / (M_PI * 2));
    cout << "bounds " << bound0 << " " << bound1 << endl;
    cout << "phase " << phase0 << ", " << phase1 << endl;
    if((int)phase0 == (int)phase1){
      // solution not found
      continue;
    }
    double n = min(phase0, phase1) + 1.0;
    double v = dtheta + n * (M_PI * 2) - M_PI;  // v = w0 t0 + w1 t1.
    // Here, w0 t0 + w1 t1 = v.
    //       t0 + t1 = remain_time.
    // Solve this.
    double t0 = (v - w1 * remain_time) / (w0 - w1);
    double t1 = remain_time - t0;

    cout << "t0, t1 = " << t0 << ", " << t1 << endl;
    assert(t0 > 0);
    assert(t1 > 0);

    start_push = ovm.current_timestep() + t0;
    again_push = start_push + tt_change_orbits;
    final_move = again_push + t1;
    break;
  }
  
  cout << "#Start push: " << start_push << endl;
}

void showEllipse(OVM& ovm) {
  double long_r, T;
  P another_focus;
  P apogee, perigee;
  double time_apogee, time_perigee;
  getellipse(true,
	     ovm,
	     long_r, T,
	     another_focus, apogee, perigee,
	     time_apogee, time_perigee);
  double short_r = sqrt(long_r * long_r - abs(0.5 * another_focus) * abs(0.5 * another_focus));
  cout << "# ellipse: " << abs(apogee)
       << " " << abs(perigee)
       << " " << T
       << " " << apogee
       << " " << perigee
       << " " << time_apogee
       << " " << time_perigee
       << endl;
}

void logging(OVM& ovm) {
#if LOG_VIS
  ovm.print_vis();
  showEllipse(ovm);
#endif
}

int main(int argc, char** argv) {
    cout.setf(ios::fixed);
    cout << setprecision(10);
    if (argc == 1) {
      printf("usage: meetgreet2 {3001|3002|3003|3004}\n");
      return 0;
    }
    OVM ovm(atoi(argv[1]));
    ovm.set_auto_save(true);

    const bool dry_run = argc >= 3 && strcmp(argv[2], "n") == 0;
    {
      ovm.run();
      ovm.print_vis();
      ovm.run();
      ovm.print_vis();
    }
    const double dir0 = (arg(ovm.current_fate_position() /
			     ovm.previous_fate_position()) > 0 ? 1 : -1);
    const double dir1 = (arg(ovm.current_target_position() /
			     ovm.previous_target_position()) > 0 ? 1 : -1);

    // estimate apogee / perigees
    double long_r0, long_r1, T0, T1;
    P af0, af1, ap0, ap1, pr0, pr1;
    double ta0, ta1, tp0, tp1;
    getellipse(true,
	       ovm,
               long_r0, T0,
               af0, ap0, pr0,
               ta0, tp0);
    getellipse(false,
	       ovm,
               long_r1, T1,
               af1, ap1, pr1,
               ta1, tp1);
    
    // wait for p0 to get perigee
    while(tp0 - ovm.current_timestep() > 0.5) {
      ovm.run();
      if (ovm.current_timestep() % 100 == 0) {
	logging(ovm);
      }
    }
    logging(ovm);

    // perigee thrust
    {
      P cur = ovm.current_fate_position();
      double r1 = abs(cur);
      double r0 = 2 * long_r0 - r1;
      cout << "r0: " << r0 << " r1: " << r1 << endl;
      // thrust: r0 -> r1
      const double dv = sqrt(G*Me/r1)*(1 - sqrt((2*r0)/(r0+r1)));

      P thrust = cur / abs(cur) * I * dir0 * dv;
      cout << "#Pushed at " << thrust << endl;
      if (dry_run)
	thrust = 0;
      ovm.run(thrust);
      logging(ovm);
    }
    // now satellite runs on circular orbital

    // move from perigee(self) to perigee(targ)
    double start_push, again_push, estimated_end;
    future_estimation(ovm, dir0, dir1,
                      pr1, tp1, T1, 
                      start_push, again_push, estimated_end);

    while (start_push - ovm.current_timestep() > 0.5) {
      ovm.run();
      if (ovm.current_timestep() % 100 == 0) {
	logging(ovm);
      }
    }
    logging(ovm);

    {
      // Start push.
      double r0 = abs(ovm.current_fate_position());
      double r1 = abs(pr1);
      {
	const double dv = sqrt(G*Me/r0)*(sqrt((2*r1)/(r0+r1)) - 1);
	P cur = ovm.current_fate_position();
	P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
	cout << "#Pushing at " << thrust << endl;
	if (dry_run)
	  thrust = P();
	ovm.run(thrust);
	logging(ovm);
      }

      while (again_push - ovm.current_timestep() > 0.5) {
	ovm.run();
	if (ovm.current_timestep() % 100 == 0) {
	  logging(ovm);
	}
      }
      logging(ovm);

      // Again push, to climb up (or down) to second circular orbital
      {
	const double dv = sqrt(G*Me/r1)*(1 - sqrt((2*r0)/(r0+r1)));

	P cur = ovm.current_fate_position();
	P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
	cout << "#Pushed at " << thrust << endl;
	if (dry_run)
	  thrust = P();
	ovm.run(thrust);
	logging(ovm);
      }
    }

    while (estimated_end - ovm.current_timestep() > 0.5) {
      ovm.run();
      if (ovm.current_timestep() % 100 == 0) {
	logging(ovm);
      }
    }
    logging(ovm);

    // Final push, to climb up to real elliptic orbital
    {
      const double r0 = abs(ovm.current_fate_position());
      const double r1 = abs(ap1);
      const double dv = sqrt(G*Me/r1)*(1 - sqrt((2*r0)/(r0+r1)));

      P cur = ovm.current_fate_position();
      P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
      cout << "#Pushed at " << thrust << endl;
      if (dry_run)
 	thrust = P();
      ovm.run(thrust);
      logging(ovm);
    }

    // Target should be near enough to the fate.
    // Final push is merged with the auto following.
    {
      // TODO: insert naturally
      auto_follow(ovm, 0);
    }

    printf("\nfinished!\n");
    printf("score = %.6f\n", ovm.current_score());

    return 0;
}
