#include <algorithm>
#include <complex>
#include <iomanip>
#include <iostream>
#include <cassert>
#include <cmath>
#include <cstring>
#include "../../include/ovm.h"

extern "C" const unsigned int TEAM_ID = 652;

using namespace std;

typedef complex<double> P;
const P I(0, 1);

const double G = 6.67428e-11;
const double Re = 6.357e6;
const double Me = 6.0e24;

P current_position() {
    return P(read_sensor(2), read_sensor(3));
}

P current_target() {
    return current_position() - P(read_sensor(4), read_sensor(5));
}

double current_fuel() {
    return read_sensor(1);
}

void set_thrust(const P& p) {
    set_thrust(p.real(), p.imag());
}

double inner_product(const P p1, const P p2) {
  return real(conj(p1) * p2);
}

double outer_product(const P p1, const P p2) {
  return imag(conj(p1) * p2);
}

const double PI=3.141592653589793238462;

// returns (d_theta + 2 * n * pi / dw) which is larger than least_ans.
double theta_div(double d_theta, double dw, double least_ans) {
  if (dw < 0) {
    d_theta = -d_theta;
    dw = -dw;
  }
  const double least_theta = dw * least_ans;
  if (d_theta > least_ans + 2 * PI) {
    double c = floor((d_theta - least_theta) / (2 * PI));
    d_theta -= c * 2 * PI;
  }
  if (d_theta < least_theta) {
    double c = ceil((least_theta - d_theta) / (2 * PI));
    d_theta += c * 2 * PI;
  }
  return d_theta / dw;
}

double run_output() {
  double res = vm_run();
  P cur = current_position();
  P target = current_target();

  if(get_timestep() % 10 == 0){
   cout << "VIS " << get_timestep()
        << " 1 " << real(cur) << " " << imag(cur) << " 0 0" << endl;
   cout << "VIS " << get_timestep()
        << " 2 " << real(target) << " " << imag(target) << " 0 0" << endl;
  }

  return res;
}

using namespace std;


/*
  Given point q and velocity v, return the shape of ellipse,
  with l / s / 
 */
void getellipse(bool myself,
                double &l, double &s, double &T,
                P &anotherfocus,
                P &apogee, P &perigee,
                double &timetoapogee, double &timetoperigee)
{
  vm_save();
  P qsim = myself ? current_position() : current_target();
  double qr = abs(qsim);
  double qrprev = qr;
  double qrpprev = qr;
  perigee = qsim;
  apogee = qsim;
  timetoapogee = -1;
  for(int i = 0; i < 3000000; ++i){
    qsim = myself ? current_position() : current_target();
    vm_run();
    
    double qr = abs(qsim);
    if(qrpprev < qrprev && qrprev > qr) {
      apogee = qsim;
      if(timetoapogee != -1){
        T = i - timetoapogee;
        break;
      }
      timetoapogee = i;
    }
    if(qrpprev > qrprev && qrprev < qr) {
      perigee = qsim;
      timetoperigee = i;
    }
    qrpprev = qrprev;
    qrprev = qr;
  }
  vm_load();
  l = abs(apogee - perigee) * 0.5;
  s = 0.0; // FIXME

  anotherfocus = apogee - (P(0., 0.) - perigee);
}

void future_estimation(double current_time,
                       double dir0, double dir1,
                       const P& target_point,
                       double target_time,
                       double target_cycle,
                       double& start_push, double& again_push, 
                       double& final_move) {
  double target_orbital = abs(target_point);
  const P start = current_position();
  //const P target = current_target();
  const double r0 = abs(start);
  const double r1 = target_orbital;

  // Time required to move from start orbit to the goal orbit.
  const double elipse_a = 0.5 * (r0 + r1);
  const double tt_change_orbits = PI * sqrt(elipse_a * elipse_a * elipse_a /  (G*Me));
  const double w0 = sqrt(G * Me / (r0 * r0 * r0)) * dir0;
  const double w1 = sqrt(G * Me / (r1 * r1 * r1)) * dir1;
  
  // Solve this.
  // theta0 + w0 t0 + w1 t1 + pi = theta1 + 2 pi n
  // where  t0 + t1 + T = Tperigee - Tcur + Telliptic k
  //        T = tt_change_orbits.
  // k is a non-negative integer, and n is an integer.
  double theta1 = arg(target_point);
  double theta0 = arg(start);
  double t0, t1;
  for(int k = 0; /* */; ++k) {
    double real_target_time = target_time + k * target_cycle;
    double remain_time = real_target_time - current_time - tt_change_orbits;
    if(remain_time < 0) continue;
    // now t0 + t1 = remain_time
    double dtheta = theta1 - theta0;
    
    double bound0 = w1 * remain_time - dtheta - M_PI;
    double phase0 = ceil(bound0 / (M_PI * 2));
    double bound1 = w0 * remain_time - dtheta - M_PI;
    double phase1 = ceil(bound1 / (M_PI * 2));
    if((int)phase0 == (int)phase1){
      // solution not found
      continue;
    }
    double n;
    if(phase0 < phase1){
      n = phase0 + 1.0;
    }else{
      n = phase0;
    }
    double v = dtheta + n * (M_PI * 2) - M_PI;
    double t0 = (v - w1 * remain_time) / (w0 - w1);
    double t1 = remain_time - t0;

    assert(t0 > 0);
    assert(t1 > 0);

    start_push = current_time + t0;
    again_push = start_push + tt_change_orbits;
    final_move = again_push + t1;
    break;
  }
  
  /*
  double d_theta;
  {
    const double c = inner_product(start, target) / r0 / r1;
    const double s = outer_product(start, target) / r0 / r1;
    cout << "# c= " << c << ", s = " << s << endl;
    if (c > 0) {
      d_theta = asin(s);
    } else {
      d_theta = PI - asin(s);
    }
  }
  */

    

  /*
    const double dw = w1 - w0;
    
    estimated_end = theta_div(PI - d_theta - w0 * tt_change_orbits, dw,
    tt_change_orbits) + current_time;
    
    cout.setf(ios::fixed);
    cout << setprecision(10);
    cout << "#Current position: " << start << ", " << target << endl;
    cout << "#Current orbit: " << r0 << endl;
    cout << "#Target orbit: " << r1 << endl;
    cout << "#Diff angle: " << d_theta << endl;
    cout << "#Angle speed: " << w0 << ", " << w1 << endl;
    cout << "#Diff angle speed: " << dw << endl;
    cout << "#Time required to change orbital:" << tt_change_orbits << endl;
    cout << "#Estimated time of rendez-vous: " << estimated_end << endl;
    
    // -1 for thruster delay.
    start_push = estimated_end - tt_change_orbits - 1;
    
  */
  cout << "#Start push: " << start_push << endl;

  assert(start_push > 0);

}

inline P simulate(P cur, P curv, int steps) {
    while(steps-- > 0) {
        P curg = -cur/abs(cur)*(G*Me/norm(cur));
        P next = cur + curv + curg / 2.0;
        P nextg = -next/abs(next)*(G*Me/norm(next));
        P nextv = curv + (curg + nextg) / 2.0;
        cur = next;
        curv = nextv;
    }
    return cur;
}

inline P velocity_from_delta(P prev, P cur) {
    double r = abs(cur);
    P g = -(cur / abs(cur)) * (G*Me)/(r*r);
    return (cur - prev + g/2.0);
}

void auto_follow() {

    printf("### start auto-following...\n");

    // calibrate
    set_thrust(P(0, 0));
    vm_run();
    P prev_me = current_position();
    P prev_target = current_target();
    vm_run();
    P cur_me = current_position();
    P cur_target = current_target();

    double score;
    for(int linger = 1; ; linger++) {
        const int T = 20;

        printf("# set linger = %d\n", linger);

        // wait for consecutive success of (linger) times
        int success = 0;
        while(success < linger) {

            P cur_me_v = velocity_from_delta(prev_me, cur_me);
            P cur_target_v = velocity_from_delta(prev_target, cur_target);

            // determine thrust with linear approx.
            P target_estimated = simulate(cur_target, cur_target_v, T);
            P desired_me_v = (target_estimated - cur_me) / (double)T;
            P thrust = desired_me_v - cur_me_v;
            set_thrust(thrust);

            printf("%7d: me=(%10.0f, %10.0f), target=(%10.0f, %10.0f), d=%.4f\n",
                   get_timestep(),
                   cur_me.real(), cur_me.imag(),
                   cur_target.real(), cur_target.imag(),
                   abs(cur_me-cur_target));

            printf("         using fuel %.3f / %.3f\n",
		   abs(thrust), current_fuel());

            double farthest = 0;
            // thrust and wait for T seconds
            for(int i = 0; i < T; i++) {
                score = vm_run();
                prev_me = cur_me;
                prev_target = cur_target;
                cur_me = current_position();
                cur_target = current_target();
                farthest = max(farthest, abs(cur_me-cur_target));
                if (score != 0)
                    goto FINISHED;
                if (i == 0)
                    set_thrust(P());
            }
            printf("           -> farthest = %.3f\n", farthest);

            // enough near to be accepted?
            if (farthest < 1e3)
                success++;
            else
                success = 0;
        }

        // hope and wait to complete
        while((score = vm_run()) == 0) {
            prev_me = cur_me;
            prev_target = cur_target;
            cur_me = current_position();
            cur_target = current_target();
            double d = abs(cur_me-cur_target);
            if (d > 1e3)
                break;
        }

        if (score != 0)
            break;
    }

FINISHED:
    printf("mission completed!\n");
    printf("score = %.6f\n", score);
    printf("# %7d: f=%9.3f\n", get_timestep(), current_fuel());
    exit(0);
}

int main(int argc, char** argv) {
    cout.setf(ios::fixed);
    cout << setprecision(10);
    if (argc == 1) {
      printf("usage: meetgreet2 {3001|3002|3003|3004}\n");
      return 0;
    }
    vm_init(atoi(argv[1]));
    const bool dry_run = argc >= 3 && strcmp(argv[2], "n") == 0;
    double dir0, dir1;
    P p0, p1, v0, v1;
    {
        run_output();
        const P p10 = current_position(); p0 = p10;
        const P p11 = current_target();   p1 = p11;
        run_output();
        const P p20 = current_position();
        const P p21 = current_target();
        double r0 = abs(p10); P g10 = - G * Me * p10 / (r0 * r0 * r0);
        double r1 = abs(p11); P g11 = - G * Me * p11 / (r1 * r1 * r1);
        v0 = p20 - p10 - 0.5 * g10; //P(get_vx(), get_vy());
        v1 = p21 - p11 - 0.5 * g11; // may contain large error
        dir0 = (arg(p20/p10) > 0 ? 1 : -1);
        dir1 = (arg(p21/p11) > 0 ? 1 : -1);
    }

    const P start = current_position();
    const P target = current_target();
    double r0 = abs(start);
    double r1 = abs(target);

    // estimate apogee / perigees
    double l0, l1, s0, s1, T0, T1;
    P af0, af1, ap0, ap1, pr0, pr1;
    double ta0, ta1, tp0, tp1;
    getellipse(true,
               l0, s0, T0,
               af0, ap0, pr0,
               ta0, tp0);
    ta0 += 2; tp0 += 2;
    getellipse(false,
               l1, s1, T1,
               af1, ap1, pr1,
               ta1, tp1);
    ta1 += 2; tp1 += 2;
    
    // wait for p0 to get perigee
    while(tp0 - get_timestep() > 1.5) {
      run_output();
    }
    // perigee thrust
    {
      P cur = current_position();
      double r1 = abs(cur);
      double r0 = 2 * l0 - r1;
      cout << "r0: " << r0 << " r1: " << r1 << endl;
      // thrust: r0 -> r1
      const double dv = sqrt(G*Me/r1)*(1 - sqrt((2*r0)/(r0+r1)));

      P thrust = cur / abs(cur) * I * dir0 * dv;
      cout << "#Pushed at " << thrust << endl;
      if (!dry_run)
        set_thrust(thrust);
    }
    run_output();
    set_thrust(P());
    // now satellite runs on circular orbital

    // move from perigee(self) to perigee(targ)
    double start_push, again_push, estimated_end;
    future_estimation(get_timestep(), dir0, dir1,
                      pr1, tp1, T1, 
                      start_push, again_push, estimated_end);

    while (start_push - get_timestep() > 1.5) {
      run_output();
    }

    // Start push.
    r0 = abs(current_position());
    r1 = abs(pr1);
    {
      const double dv = sqrt(G*Me/r0)*(sqrt((2*r1)/(r0+r1)) - 1);
      P cur = current_position();
      P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
      cout << "#Pushing at " << thrust << endl;
      if (!dry_run)
        set_thrust(thrust);
    }
    run_output();
    set_thrust(P());

    while (again_push - get_timestep() > 1.5) {
      run_output();
    }

    // Again push, to climb up (or down) to second circular orbital
    {
      const double dv = sqrt(G*Me/r1)*(1 - sqrt((2*r0)/(r0+r1)));

      P cur = current_position();
      P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
      cout << "#Pushed at " << thrust << endl;
      if (!dry_run)
        set_thrust(thrust);
    }
    run_output();
    set_thrust(P());

    while (estimated_end - get_timestep() > 1.5) {
      run_output();
    }

    // Final push, to climb up to real elliptic orbital
    {
      r0 = r1;
      r1 = abs(ap1);
      const double dv = sqrt(G*Me/r1)*(1 - sqrt((2*r0)/(r0+r1)));

      P cur = current_position();
      P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
      cout << "#Pushed at " << thrust << endl;
      if (!dry_run)
        set_thrust(thrust);
    }
    run_output();
    set_thrust(P());

    {
        // TODO: insert naturally
        vm_run();
        auto_follow();
    }

    cout << "# waiting to complete the mission..." << endl;
    double score;
    {
        double mini = 1e100, maxi = 0;
        int time_limit = 9000;
        do {
            if (--time_limit < 0) {
	        cout << "#Time limit: exitting." << endl;
	        break;
	    }

            P cur = current_position();
            double d = abs(cur);
            mini = min(mini, d);
            maxi = max(maxi, d);
            printf("# %7d: (%10.0f, %10.0f) d=%10.0f [%10.0f..%10.0f] f=%9.3f\n", get_timestep(), cur.real(), cur.imag(), d, mini, maxi, current_fuel());
        } while((score = run_output()) == 0);
    }
    printf("\nfinished!\n");
    printf("score = %.6f\n", score);

    return 0;
}


