#include <iostream>
#include <complex>
#include <cstdlib>
#include "ovm.h"

extern "C" const unsigned int TEAM_ID = 652;

using namespace std;

typedef complex<double> P;
const P I(0, 1);

const double G = 6.67428e-11;
const double Re = 6.357e6;
const double Me = 6.0e24;

P current_position() {
    return P(read_sensor(2), read_sensor(3));
}

double current_fuel() {
    return read_sensor(1);
}

void set_thrust(const P& p) {
    set_thrust(p.real(), p.imag());
}

P current_velocity() {
    return P(get_vx(), get_vy());
}


int main(int argc, char** argv) {

    if (argc == 1) {
        printf("usage: hohmann 100[1234]\n");
        return 0;
    }
    vm_init(atoi(argv[1]));

    double dir;
    {
        vm_run();
        const P p1 = current_position();
        vm_run();
        const P p2 = current_position();
        dir = (arg(p2/p1) > 0 ? 1 : -1);
    }

    const double R = read_sensor(4);

    double excess_use;
    { // [これはひどい]
        const double r = abs(current_position());
        double estimaed_usage = sqrt(G*Me/r)*(sqrt((2*R)/(r+R)) - 1) + sqrt(G*Me/R)*(1 - sqrt((2*r)/(r+R)));
        excess_use = current_fuel() - estimaed_usage - 50; // 50くらいのこす
    }

    {
        P p = current_position();
        set_thrust((excess_use/2)*p/abs(p));
        vm_run();
        set_thrust(-(excess_use/2)*p/abs(p));
        vm_run();
    }

    const P start = current_position();
    const double r = abs(start);

    printf("current orbit: r = %.0f\n", r);
    printf("target orbit: R = %.0f\n", R);

    {
        double v = sqrt(G*Me/r)*(sqrt((2*R)/(r+R)) - 1);
        P cur = current_position();
        P curv = current_velocity();
        P thrust = (cur/abs(cur)*I) * v * dir;
        printf("start transition at: (%.0f, %.0f) d=%.0f v=(%.0f, %.0f)\n", cur.real(), cur.imag(), abs(cur), curv.real(), curv.imag());
        printf("set thrust: (%.0f, %.0f)\n", thrust.real(), thrust.imag());
        set_thrust(thrust);
    }

    vm_run();
    set_thrust(P());

    printf("waiting to arrive the target orbit...\n");
    {
        double ld = abs(current_position());
        double mini = ld, maxi = ld;
        for(;;) {
            P cur = current_position();
            P curv = current_velocity();
            double d = abs(cur);
            mini = min(mini, d);
            maxi = max(maxi, d);
            printf("%7d: (%10.0f, %10.0f) v=(%5.0f, %5.0f)|%5.0f| d=%10.0f [%10.0f..%10.0f] f=%9.3f\r", get_timestep(), cur.real(), cur.imag(), curv.real(), curv.imag(), abs(curv), d, mini, maxi, current_fuel());
            if ((R > r && d >= R-1e2) || (R < r && d <= R+1e2))
                break;
            ld = d;
            vm_run();
        }
    }
    printf("\narrived!\n");
    
    /*
    {
        double v = sqrt(G*Me/R)*(1 - sqrt((2*r)/(r+R)));
        P cur = current_position();
        P curv = current_velocity();
        P thrust = (cur/abs(cur)*I) * v * dir;
        printf("current velocity: (%.0f, %.0f)\n", curv.real(), curv.imag());
        printf("end transition at: (%.0f, %.0f) d=%.0f\n", cur.real(), cur.imag(), abs(cur));
        printf("set thrust: (%.0f, %.0f)\n", thrust.real(), thrust.imag());
        set_thrust(thrust);
    }
    /*/
    {
        // try to compute thrust exactly...
        P cur = current_position();
        P curv = current_velocity();
        P expv = (cur / abs(cur)) * I * dir * sqrt(G*Me/abs(cur));
        P thrust = expv - curv;
        printf("end transition at: (%.0f, %.0f) d=%.0f\n", cur.real(), cur.imag(), abs(cur));
        printf("current velocity: (%.0f, %.0f)\n", curv.real(), curv.imag());
        printf("orbit velocity: (%.0f, %.0f)\n", expv.real(), expv.imag());
        printf("set thrust: (%.0f, %.0f)\n", thrust.real(), thrust.imag());
        set_thrust(thrust);
    }
    //*/

    vm_run();
    set_thrust(P());

    printf("waiting to complete the mission...\n");
    double score;
    {
        double mini = 1e100, maxi = 0;
        do {
            P cur = current_position();
            P curv = current_velocity();
            double d = abs(cur);
            mini = min(mini, d);
            maxi = max(maxi, d);
            printf("%7d: (%10.0f, %10.0f) v=(%5.0f, %5.0f)|%5.0f| d=%10.0f [%10.0f..%10.0f] f=%9.3f\r", get_timestep(), cur.real(), cur.imag(), curv.real(), curv.imag(), abs(curv), d, mini, maxi, current_fuel());
        } while((score = vm_run()) == 0);
    }
    printf("\nfinished!\n");
    printf("score = %.6f\n", score);
    printf("furl = %.3f\n", current_fuel());

    return 0;
}
