#include <algorithm>
#include <complex>
#include <iomanip>
#include <iostream>
#include <cassert>
#include <cmath>
#include <cstring>
#include "ovm.hpp"
#include "../clearsky/autofollow.hpp"

using namespace std;

double inner_product(const P p1, const P p2) {
  return real(conj(p1) * p2);
}

double outer_product(const P p1, const P p2) {
  return imag(conj(p1) * p2);
}

bool is_counterclockwise(const P &q, const P &p)
{
  return (outer_product(q, p) > 0.);
}

double get_direction(const P &q, const P &p)
{
  return is_counterclockwise(q, p) > 0 ? 1. : -1.;
}

double get_omega(P q)
{
  double r = abs(q);
  double w = sqrt(G * Me / (r * r * r));
  return w;
}

double get_omega(P q, P p)
{
  double r = abs(q);
  double w = sqrt(G * Me / (r * r * r)) * 
    (is_counterclockwise(q, p) ? 1. : -1.);
  return w;
}

void run_with_log(OVM &ovm)
{
  ovm.run();
  if(ovm.current_timestep() % 200 == 0)
    ovm.print_vis();
}

// returns (d_theta + 2 * n * pi / dw) which is larger than least_ans.
double theta_div(double d_theta, double dw, double least_ans) {
  if (dw < 0) {
    d_theta = -d_theta;
    dw = -dw;
  }
  const double least_theta = dw * least_ans;
  if (d_theta > least_ans + 2 * M_PI) {
    double c = floor((d_theta - least_theta) / (2 * M_PI));
    d_theta -= c * 2 * M_PI;
  }
  if (d_theta < least_theta) {
    double c = ceil((least_theta - d_theta) / (2 * M_PI));
    d_theta += c * 2 * M_PI;
  }
  return d_theta / dw;
}

void future_estimation(OVM& ovm, double& start_push, double& estimated_end) {
  const P start = ovm.current_fate_position();
  const P target = ovm.current_target_position();
  const double r0 = abs(start);
  const double r1 = abs(target);

  double delta_theta = arg(target / start);

  // Time required to move from start orbit to the goal orbit.
  const double elipse_a = 0.5 * (r0 + r1);
  const double tt_change_orbits =
    M_PI * sqrt(elipse_a * elipse_a * elipse_a / (G*Me));

  // Solve this.
  // theta0 + w0(t - T) + pi = theta1 + w1 * t,
  // where T = tt_change_orbits.
  // <=>  d_theta + d_w * t + w0 * T - pi = 0
  // t = (PI - d_theta - w0*T + 2nPi) / dw.  where n is int.
  const double w0 = get_omega(start, ovm.current_fate_velocity());
  const double w1 = get_omega(target, ovm.current_target_velocity());

  const double dw = w1 - w0;
  estimated_end = theta_div(M_PI - delta_theta - w0 * tt_change_orbits, dw,
			    tt_change_orbits) + ovm.current_timestep();
  // -1 for thruster delay.
  start_push = estimated_end - tt_change_orbits;

  cout.setf(ios::fixed);
  cout << setprecision(10);
  cout << "#Current position: " << start << ", " << target << endl;
  cout << "#Current orbit: " << r0 << endl;
  cout << "#Target orbit: " << r1 << endl;
  cout << "#Diff angle: " << delta_theta << endl;
  cout << "#Angle speed: " << w0 << ", " << w1 << endl;
  cout << "#Diff angle speed: " << dw << endl;
  cout << "#Time required to change orbital:" << tt_change_orbits << endl;
  cout << "#Estimated time of rendez-vous: " << estimated_end << endl;
  cout << "#Start push: " << start_push << endl;
}

void bump(OVM &ovm, double r0, double r1)
{
  const double dv = sqrt(G*Me/r0)*(sqrt((2*r1)/(r0+r1)) - 1);
  if (dv > ovm.current_fuel()) {
    // If the fuel is too much, do nothing.
    return;
  }

  P cur = ovm.current_fate_position();
  double dir0 = is_counterclockwise(cur, ovm.current_fate_velocity()) ? 1 : -1;
  P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
  cout << "#Pushed at " << thrust << " time = " << ovm.current_timestep() << endl;

  if (abs(thrust) > ovm.current_fuel()) {
    thrust = P();
  }

  ovm.print_vis();
  ovm.run(thrust);
  ovm.print_vis();
}

void arrive(OVM &ovm, double r0, double r1)
{
  const double dv = sqrt(G*Me/r1)*(1 - sqrt((2*r0)/(r0+r1)));
  if (dv > ovm.current_fuel()) {
    // If the fuel is too much, do nothing.
    return;
  }

  P cur = ovm.current_fate_position();
  // FIXME: pass direction for different direction

  double dir0 = is_counterclockwise(cur, ovm.current_fate_velocity()) ? 1 : -1;
  P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
  cout << "#Pushed at " << thrust << " time = " << ovm.current_timestep() << endl;

  if (abs(thrust) > ovm.current_fuel()) {
    thrust = P();
  }

  ovm.print_vis();
  ovm.run(thrust);
  ovm.print_vis();
}


int main(int argc, char** argv) {
  cout.setf(ios::fixed);
  cout << setprecision(10);
  if (argc == 1) {
    printf("usage: rendez-vous 200[1234]\n");
    return 0;
  }

  OVM ovm(atoi(argv[1]));
  ovm.set_auto_save(true);

  ovm.run();
  ovm.run();

  double start_push, estimated_end;
  future_estimation(ovm, start_push, estimated_end);

  while (start_push - ovm.current_timestep() > 0.5) {
    ovm.run();
  }

  const double r0 = abs(ovm.current_fate_position());
  const double r1 = abs(ovm.current_target_position());

  // Start push.
  bump(ovm, r0, r1);

  while (estimated_end - ovm.current_timestep() > 1.5) {
    ovm.run();
  }

  // Finish push.
  arrive(ovm, r0, r1);
  {
    // TODO: insert naturally
    auto_follow(ovm, 0);
  }

  printf("\nfinished!\n");
  printf("score = %.6f\n", ovm.current_score());
  return 0;
}
