#include <algorithm>
#include <complex>
#include <iomanip>
#include <iostream>
#include <cassert>
#include <cmath>
#include <cstring>
#include "ovm.h"

extern "C" const unsigned int TEAM_ID = 652;

using namespace std;

typedef complex<double> P;
const P I(0, 1);

const double G = 6.67428e-11;
const double Re = 6.357e6;
const double Me = 6.0e24;

P current_position() {
    return P(read_sensor(2), read_sensor(3));
}

P current_target() {
    return current_position() - P(read_sensor(4), read_sensor(5));
}

double current_fuel() {
    return read_sensor(1);
}

void set_thrust(const P& p) {
    set_thrust(p.real(), p.imag());
}

double inner_product(const P p1, const P p2) {
  return real(conj(p1) * p2);
}

double outer_product(const P p1, const P p2) {
  return imag(conj(p1) * p2);
}

const double PI=3.141592653589793238462;

// returns (d_theta + 2 * n * pi / dw) which is larger than least_ans.
double theta_div(double d_theta, double dw, double least_ans) {
  if (dw < 0) {
    d_theta = -d_theta;
    dw = -dw;
  }
  const double least_theta = dw * least_ans;
  if (d_theta > least_ans + 2 * PI) {
    double c = floor((d_theta - least_theta) / (2 * PI));
    d_theta -= c * 2 * PI;
  }
  if (d_theta < least_theta) {
    double c = ceil((least_theta - d_theta) / (2 * PI));
    d_theta += c * 2 * PI;
  }
  return d_theta / dw;
}

double run_output() {
  double res = vm_run();
  P cur = current_position();
  P target = current_target();

/*
   cout << "VIS " << get_timestep()
        << " 1 " << real(cur) << " " << imag(cur) << " 0 0" << endl;
   cout << "VIS " << get_timestep()
        << " 2 " << real(target) << " " << imag(target) << " 0 0" << endl;
*/

  return res;
}

void future_estimation(double current_time,
		       double dir0, double dir1,
		       double& start_push, double& estimated_end) {
  const P start = current_position();
  const P target = current_target();
  const double r0 = abs(start);
  const double r1 = abs(target);

  double d_theta;
  {
    const double c = inner_product(start, target) / r0 / r1;
    const double s = outer_product(start, target) / r0 / r1;
    cout << "# c= " << c << ", s = " << s << endl;
    if (c > 0) {
      d_theta = asin(s);
    } else {
      d_theta = PI - asin(s);
    }
  }
  const double w0 = sqrt(G * Me / (r0 * r0 * r0)) * dir0;
  const double w1 = sqrt(G * Me / (r1 * r1 * r1)) * dir1;

  // Time required to move from start orbit to the goal orbit.
  const double elipse_a = 0.5 * (r0 + r1);
  const double tt_change_orbits = PI * sqrt(elipse_a * elipse_a * elipse_a /
					    (G*Me));
    

  // Solve this.
  // theta0 + w0(t - T) + pi = theta1 + w1 * t,
  // where T = tt_change_orbits.
  // <=>  d_theta + d_w * t + w0 * T - pi = 0
  // t = (PI - d_theta - w0*T + 2nPi) / dw.  where n is int.

  const double dw = w1 - w0;

  estimated_end = theta_div(PI - d_theta - w0 * tt_change_orbits, dw,
			    tt_change_orbits) + current_time;

  cout.setf(ios::fixed);
  cout << setprecision(10);
  cout << "#Current position: " << start << ", " << target << endl;
  cout << "#Current orbit: " << r0 << endl;
  cout << "#Target orbit: " << r1 << endl;
  cout << "#Diff angle: " << d_theta << endl;
  cout << "#Angle speed: " << w0 << ", " << w1 << endl;
  cout << "#Diff angle speed: " << dw << endl;
  cout << "#Time required to change orbital:" << tt_change_orbits << endl;
  cout << "#Estimated time of rendez-vous: " << estimated_end << endl;

  // -1 for thruster delay.
  start_push = estimated_end - tt_change_orbits - 1;

  cout << "#Start push: " << start_push << endl;

  assert(start_push > 0);

}

inline P simulate(P cur, P curv, int steps) {
    while(steps-- > 0) {
        P curg = -cur/abs(cur)*(G*Me/norm(cur));
        P next = cur + curv + curg / 2.0;
        P nextg = -next/abs(next)*(G*Me/norm(next));
        P nextv = curv + (curg + nextg) / 2.0;
        cur = next;
        curv = nextv;
    }
    return cur;
}

inline P velocity_from_delta(P prev, P cur) {
    double r = abs(cur);
    P g = -(cur / abs(cur)) * (G*Me)/(r*r);
    return (cur - prev + g/2.0);
}

void auto_follow() {

    printf("### start auto-following...\n");

    // calibrate
    vm_run();
    set_thrust(P(0, 0));
    P prev_me = current_position();
    P prev_target = current_target();
    vm_run();
    P cur_me = current_position();
    P cur_target = current_target();

    double score;
    for(int linger = 1; ; linger++) {
        const int T = 12;

        printf("# set linger = %d\n", linger);

        // wait for consecutive success of (linger) times
        int success = 0;
        while(success < linger) {

            P cur_me_v = velocity_from_delta(prev_me, cur_me);
            P cur_target_v = velocity_from_delta(prev_target, cur_target);

            // determine thrust with linear approx.
            P target_estimated = simulate(cur_target, cur_target_v, T);
            P desired_me_v = (target_estimated - cur_me) / (double)T;
            P thrust = desired_me_v - cur_me_v;
            set_thrust(thrust);

            printf("%7d: me=(%10.0f, %10.0f), target=(%10.0f, %10.0f), d=%.4f\n",
                   get_timestep(),
                   cur_me.real(), cur_me.imag(),
                   cur_target.real(), cur_target.imag(),
                   abs(cur_me-cur_target));

            printf("         using fuel %.3f / %.3f\n",
		   abs(thrust), current_fuel());

            double farthest = 0;
            // thrust and wait for T seconds
            for(int i = 0; i < T; i++) {
                score = vm_run();
                prev_me = cur_me;
                prev_target = cur_target;
                cur_me = current_position();
                cur_target = current_target();
                farthest = max(farthest, abs(cur_me-cur_target));
                if (score != 0)
                    goto FINISHED;
                if (i == 0)
                    set_thrust(P());
            }
            printf("           -> farthest = %.3f\n", farthest);

            // enough near to be accepted?
            if (farthest < 1e3)
                success++;
            else
                success = 0;
        }

        // hope and wait to complete
        while((score = vm_run()) == 0) {
            prev_me = cur_me;
            prev_target = cur_target;
            cur_me = current_position();
            cur_target = current_target();
            double d = abs(cur_me-cur_target);
            if (d > 1e3)
                break;
        }

        if (score != 0)
            break;
    }

FINISHED:
    printf("mission completed!\n");
    printf("score = %.6f\n", score);
    printf("# %7d: f=%9.3f\n", get_timestep(), current_fuel());
    exit(0);
}

int main(int argc, char** argv) {
    cout.setf(ios::fixed);
    cout << setprecision(10);
    if (argc == 1) {
      printf("usage: rendez-vous 200[1234]\n");
      return 0;
    }
    vm_init(atoi(argv[1]));
    const bool dry_run = argc >= 3 && strcmp(argv[2], "n") == 0;
    double dir0, dir1;
    {
        run_output();
        const P p10 = current_position();
	const P p11 = current_target();
        run_output();
        const P p20 = current_position();
	const P p21 = current_target();
        dir0 = (arg(p20/p10) > 0 ? 1 : -1);
        dir1 = (arg(p21/p11) > 0 ? 1 : -1);
    }

    const P start = current_position();
    const P target = current_target();
    const double r0 = abs(start);
    const double r1 = abs(target);

    double start_push, estimated_end;
    future_estimation(get_timestep(), dir0, dir1, start_push, estimated_end);

    while (start_push - get_timestep() > 0.5) {
      run_output();
    }

    // Start push.
    {
      const double dv = sqrt(G*Me/r0)*(sqrt((2*r1)/(r0+r1)) - 1);
      P cur = current_position();
      P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
      cout << "#Pushing at " << thrust << endl;
      if (!dry_run)
	set_thrust(thrust);
    }
    run_output();
    set_thrust(P());

    while (estimated_end - get_timestep() > 1.5) {
      run_output();
    }

    // Finish push.
    {
      const double dv = sqrt(G*Me/r1)*(1 - sqrt((2*r0)/(r0+r1)));

      P cur = current_position();
      P thrust = (cur / abs(cur) * P(0, 1)) * dv * dir0;
      cout << "#Pushed at " << thrust << endl;
      if (!dry_run)
	set_thrust(thrust);
    }

    {
        // TODO: insert naturally
        auto_follow();
    }

    cout << "# waiting to complete the mission..." << endl;
    double score;
    {
        double mini = 1e100, maxi = 0;
	int time_limit = 9000;
        do {
            if (--time_limit < 0) {
	        cout << "#Time limit: exitting." << endl;
	        break;
	    }

            P cur = current_position();
            double d = abs(cur);
            mini = min(mini, d);
            maxi = max(maxi, d);
            printf("# %7d: (%10.0f, %10.0f) d=%10.0f [%10.0f..%10.0f] f=%9.3f\n", get_timestep(), cur.real(), cur.imag(), d, mini, maxi, current_fuel());
        } while((score = run_output()) == 0);
    }
    printf("\nfinished!\n");
    printf("score = %.6f\n", score);


    return 0;
}
