package jp.dip.icfp2009;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;

public class VisualFateDocument {

	public static final double EARTH_RADIUS = 6.357e6;

	public static class State {
		private int time;
		private double x;
		private double y;
		private double vx;
		private double vy;
		public State(int time, double x, double y, double vx, double vy) {
			this.time = time;
			this.x = x;
			this.y = y;
			this.vx = vx;
			this.vy = vy;
		}

		public int getTime() {
			return time;
		}
		
		public double getX() {
			return x;
		}
		
		public double getY() {
			return y;
		}
	
		public double getVx() {
			return vx;
		}
		
		public double getVy() {
			return vy;
		}
	}

	private static final Comparator<State> STATE_TIME_COMPARATOR = new Comparator<State>() {
		@Override
		public int compare(State o1, State o2) {
			if (o1.getTime() < o2.getTime()) {
				return -1;
			}
			if (o1.getTime() > o2.getTime()) {
				return 1;
			}
			return 0;
		}
	};

	public static class Entity {

		private String name;

		private List<State> stateList;
		
		public Entity(String name) {
			if (name == null) {
				throw new NullPointerException("Name is null.");
			}
			
			this.name = name;
			this.stateList = new ArrayList<State>(3000000);
		}

		public String getName() {
			return name;
		}

		public void addState(int time, double x, double y, double vx, double vy) {
			State state = new State(time, x, y, vx, vy);
			if (!stateList.isEmpty()) {
				int lastIndex = stateList.size() - 1;
				int lastTime = stateList.get(lastIndex).getTime();
				if (lastTime == time) {
					stateList.set(lastIndex, state);
					return;
				}
				if (lastTime < time) {
					stateList.add(state);
					return;
				}
			}
			
			int index = Collections.binarySearch(stateList, state, STATE_TIME_COMPARATOR);
			if (index >= 0) {
				// Overwrite.
				stateList.set(index, state);
			} else {
				index = -index-1;
				stateList.add(index, state);
			}
		}

		public State getStateAt(int time) {
			State keyState = new State(time, 0, 0, 0, 0);
			
			int index = Collections.binarySearch(stateList, keyState, STATE_TIME_COMPARATOR);
			if (index < 0) {
				index = -index-1;
			}
			if (index == stateList.size()) {
				--index;
			}
			return stateList.get(index);
		}
		
		public List<State> getStateList() {
			return Collections.unmodifiableList(stateList);
		}
	}

	private Map<String, Entity> entityMap;
	private int maxTime;
	
	public VisualFateDocument() {
		this.entityMap = Collections.emptyMap();
		this.maxTime = 0;
		this.currentTime = 0;
		this.scale = EARTH_RADIUS / 1000 * 10 * 2;
	}
	

	public void loadFile(File file) throws IOException, FileNotFoundException {
		BufferedReader reader = new BufferedReader(new FileReader(file));
		Map<String, Entity> newEntityMap = new HashMap<String, Entity>();
		
		// format: VIS (time) (name) (x) (y) (vx) (vy)
		int maxTime = 0;
		while (true) {
			String line = reader.readLine();
			if (line == null) {
				break;
			}

			if (!line.startsWith("VIS")) {
				continue;
			}
			
			Scanner lineScanner = new Scanner(line);
			if (!lineScanner.next().equals("VIS")) {
				continue;
			}

			int time = lineScanner.nextInt();
			if (maxTime < time) {
				maxTime = time;
			}

			String name = lineScanner.next();
			double x = lineScanner.nextDouble();
			double y = lineScanner.nextDouble();
			double vx = lineScanner.nextDouble();
			double vy = lineScanner.nextDouble();

			if (lineScanner.ioException() != null) {
				throw lineScanner.ioException();
			}
		
			Entity entity = newEntityMap.get(name);
			if (entity == null) {
				entity = new Entity(name);
				newEntityMap.put(name, entity);
			}
			entity.addState(time, x, y, vx, vy);
		}
		
		this.maxTime = maxTime;
		this.entityMap = Collections.unmodifiableMap(newEntityMap);
	}

	public int getMaxTime() {
		return maxTime;
	}

	public void setMaxTime(int maxTime) {
		if (maxTime < 0) {
			throw new IllegalArgumentException("maxTime is negative: " + maxTime);
		}
		this.maxTime = maxTime;
	}
	
	public Map<String, Entity> getEntityMap() {
		return entityMap;
	}

	public void setEntityMap(Map<String, Entity> newEntityMap) {
		if (newEntityMap == null) {
			throw new NullPointerException("newEntityMap is null.");
		}
		this.entityMap = newEntityMap;
	}
	
	// Followings are state.
	private int currentTime;

	public void setCurrentTime(int currentTime) {
		// Clip.
		if (currentTime < 0) {
			currentTime = 0;
		}
		if (currentTime > maxTime) {
			currentTime = maxTime;
		}

		this.currentTime = currentTime;
	}

	public int getCurrentTime() {
		return currentTime;
	}

	private double scale;
	
	public void setScale(double scale) {
		this.scale = scale;
	}

	public double getScale() {
		return scale;
	}	

	private double centerX;
	private double centerY;
	
	public double getCenterX() {
		return centerX;
	}

	public double getCenterY() {
		return centerY;
	}

	public void setCenter(double x, double y) {
		this.centerX = x;
		this.centerY = y;
	}
}
