package jp.dip.icfp2009;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.channels.FileChannel;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import javax.swing.JFileChooser;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JProgressBar;
import javax.swing.JSlider;
import javax.swing.SwingWorker;

import jp.dip.icfp2009.VisualFateDocument.Entity;

@SuppressWarnings("serial")
public class VisualFateMenu extends JMenuBar {

	private static class DocumentLoader extends SwingWorker<Void, Void> {
		private final VisualFateFrame frame;
		private final VisualFateLoadDialog dialog;
		private final VisualFateDocument document;
		private final File file;
		private Map<String, Entity> newEntityMap;
		private int maxTime;
		
		private DocumentLoader(VisualFateFrame frame, VisualFateLoadDialog dialog, VisualFateDocument document, File file) {
			if (frame == null) {
				throw new NullPointerException("frame is null.");
			}
			if (dialog == null) {
				throw new NullPointerException("dialog is null.");
			}
			if (document == null) {
				throw new NullPointerException("document is null.");
			}
			if (file == null) {
				throw new NullPointerException("File is null.");
			}
			this.frame = frame;
			this.dialog = dialog;
			this.document = document;
			this.file = file;
			this.newEntityMap = null;
			this.maxTime = 0;
			dialog.getProgressBar().setValue(0);
			dialog.repaint();
			this.addPropertyChangeListener(new ProgressListener(dialog.getProgressBar()));
		}
		
		@Override
		protected Void doInBackground() throws Exception {
			// format: VIS (time) (name) (x) (y) (vx) (vy)
			BufferedReader reader = null;
			Map<String, Entity> newEntityMap = new HashMap<String, Entity>();
			int maxTime = 0;
			int progress = 0;
			try {
				FileInputStream fileInputStream = new FileInputStream(this.file);
				FileChannel channel = fileInputStream.getChannel();
				reader = new BufferedReader(new InputStreamReader(fileInputStream));

				long fileSize = file.length();
				while (!isCancelled()) {
					String line = reader.readLine();
					if (line == null) {
						break;
					}
					long position = channel.position();
					int currentProgress = (int)(position * 100 / fileSize);
					if (currentProgress != progress) {
						this.setProgress(currentProgress);
					}
					
					if (!line.startsWith("VIS")) {
						continue;
					}

					StringTokenizer st = new StringTokenizer(line);
					if (!st.nextToken().equals("VIS")) {
						continue;
					}

					int time = Integer.parseInt(st.nextToken());
					if (maxTime < time) {
						maxTime = time;
					}

					String name = st.nextToken();
					double x = Double.parseDouble(st.nextToken());
					double y = Double.parseDouble(st.nextToken());
					double vx = Double.parseDouble(st.nextToken());
					double vy = Double.parseDouble(st.nextToken());

					Entity entity = newEntityMap.get(name);
					if (entity == null) {
						entity = new Entity(name);
						newEntityMap.put(name, entity);
					}
					entity.addState(time, x, y, vx, vy);
				}
					
				this.maxTime = maxTime;
				this.newEntityMap = Collections.unmodifiableMap(newEntityMap);
			} catch (IOException e) {
				if (reader != null) {
					try {
						reader.close();
					} catch (IOException ignore) {
					}
				}
			}
			if (reader != null) {
				reader.close();
			}
			return null;
		}

		@Override
		protected void process(List<Void> unused) {
			dialog.getProgressBar().setValue(this.getProgress());
			dialog.repaint();
		}
		
		@Override
		protected void done() {
			frame.setEnabled(true);
			if (!isCancelled()) {
				try {
					get();
					this.document.setEntityMap(this.newEntityMap);
					this.document.setMaxTime(this.maxTime);
				
					JSlider timeSlider = frame.getSliderPanel().getTimePanel().getTimeSlider();
					timeSlider.setMaximum(document.getMaxTime());
					timeSlider.setValue(0);

					JSlider scaleSlider = frame.getSliderPanel().getScalePanel().getScaleSlider();
					scaleSlider.setValue(30);
				} catch (Exception e) {
					if (e instanceof FileNotFoundException) {
						FileNotFoundException exn = FileNotFoundException.class.cast(e);
						JOptionPane.showMessageDialog(
								frame, file.getAbsolutePath() + " cannot be opened: " + exn.getLocalizedMessage(),
								"Error!!!", JOptionPane.ERROR_MESSAGE);	
					} else {
						JOptionPane.showMessageDialog(
								frame, "Failed to read log: " + e.getLocalizedMessage(),
								"Error!!!", JOptionPane.ERROR_MESSAGE);
					}
				}
			}
			dialog.setVisible(false);
			frame.repaint();
		}

		private static class ProgressListener implements PropertyChangeListener {
		    private final JProgressBar progressBar;
		    private ProgressListener(JProgressBar progressBar) {
		        this.progressBar = progressBar;
		        this.progressBar.setValue(0);
		    }
		    @Override
		    public void propertyChange(PropertyChangeEvent evt) {
		    	String strPropertyName = evt.getPropertyName();
		        if ("progress".equals(strPropertyName)) {
		            progressBar.setIndeterminate(false);
		            int progress = (Integer)evt.getNewValue();
		            progressBar.setValue(progress);
		        }
		    }
		}
	}
	
	private static class FileMenu extends JMenu {
		public FileMenu(VisualFateFrame frame, VisualFateDocument document) {
			super("File");
			super.add(new OpenItem(frame, document));
			super.addSeparator();
			super.add(new ExitItem());
		}

		private static class OpenItem extends JMenuItem {
			public OpenItem(VisualFateFrame frame, VisualFateDocument document) {
				super("Open");
				super.addActionListener(new OpenListener(frame, document));
			}
			
			private static class OpenListener implements ActionListener {
				private final JFileChooser fileChooser;
				private final VisualFateFrame frame;
				private final VisualFateDocument document;
				private final VisualFateLoadDialog dialog;
				
				public OpenListener(VisualFateFrame frame, VisualFateDocument document) {
					if (frame == null) {
						throw new NullPointerException("frame is null.");
					}
					if (document == null) {
						throw new NullPointerException("document is null.");
					}
					this.fileChooser = new JFileChooser();
					this.frame = frame;
					this.document = document;
					this.dialog = new VisualFateLoadDialog(frame, document);
				}
				
				@Override
				public void actionPerformed(ActionEvent e) {
					int selected = fileChooser.showOpenDialog(frame);
					if (selected == JFileChooser.APPROVE_OPTION) {
						File file = fileChooser.getSelectedFile();
						frame.setEnabled(false);
						DocumentLoader loader = new DocumentLoader(frame, dialog, document, file);
						loader.execute();
						int centerX = frame.getWidth() / 2;
						int centerY = frame.getHeight() / 2;
						
						dialog.setLocation(centerX - dialog.getWidth() / 2,
										   centerY - dialog.getHeight() / 2);
						dialog.getProgressBar().setValue(0);
						dialog.setVisible(true);
					}
				}
			}
		}
		
		private static class ExitItem extends JMenuItem {
			public ExitItem() {
				super("EXIT");
				super.addActionListener(new ExitListener());
			}

			private static class ExitListener implements ActionListener {
				@Override
				public void actionPerformed(ActionEvent e) {
					System.exit(0);
				}
			}
		}
	}

	public VisualFateMenu(VisualFateFrame frame, VisualFateDocument document) {
		super.add(new FileMenu(frame, document));
	}
}
